/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

This source code is partially based on the zlib package by Jean-loup Gailly
and Mark Adler.  Licensing information for the original zlib package now
follows (note that this license applies to the zlib source code, not the
Compression class):

  Copyright (C) 1995-1998 Jean-loup Gailly and Mark Adler

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Jean-loup Gailly        Mark Adler
  jloup@gzip.org          madler@alumni.caltech.edu

******************************************************************************

###CLASS###
Name:      Compression
Version:   1.0
Status:    Stable
Date:      April 2004
ID:        ID_COMPRESSION
Category:  Data
Author:    Rocklyte Systems, based on work from Jean-loup Gailly and Mark Adler.
Copyright: Rocklyte Systems (c) 2001-2004.  All rights reserved.
Keywords:  compression, decompress, compress, pack, shrink, inflate, deflate

###DESCRIPTION###
<p>The Compression class provides the necessary means to compress and
decompress your data.  It provides support for file based compression
as well as memory based compression routines.  The base class uses zip
algorithms to support pkzip files, while other forms of compressed data can
be supported by installing extra compression sub-classes.</p>

<p>The Compression class is compatible with the Dynamic Markup Language.
Here are some examples that demonstrate the use of compression objects in
DML:</p>

<pre>
  &lt;!-- Extract all data from a zip file to the athene: directory --&gt;
  &lt;compression src="athene:data.zip"&gt;
    &lt;action method="decompressfile" &path="*" &destination="athene:"/&gt;
  &lt;/compression&gt;

  &lt;!-- Delete the readme.txt file from a zip file --&gt;
  &lt;compression src="athene:data.zip"&gt;
    &lt;action method="removefile" &path="readme.txt"/&gt;
  &lt;/compression&gt;

  &lt;!-- Add the readme.txt file back to the zip file --&gt;
  &lt;compression src="athene:data.zip"&gt;
    &lt;action method="compressfile" &location="readme.txt" &path=""/&gt;
  &lt;/compression&gt;
</pre>

<p>If data is to be encyrpted or decrypted then you will need to set the
<a href="#Password">Password</a> field so that it refers to a null-terminated
encryption string.  If you don't know the password for an encrypted file,
errors will be returned when trying to decompress the information.</p>

<p>To peruse the contents of a compressed file, use the
<a href="#FileList">FileList</a> field to retrieve a list of files available
in the object.  This will give you information on the name, time stamp and
size of each individual file.</p>

<p>To adjust the level of compression used to pack each file, set the
<a href="#CompressionLevel">CompressionLevel</a> field to a value between 0
and 100%.</p>

###END###

*****************************************************************************/

#include "zlib/zlib.h"
#include "zlib/zutil.h"

#include <pandora/system/all.h>
#include <pandora/files/file.h>
#include <pandora/data/compression.h>
#include <pandora/misc/time.h>
#include <pandora/modules/filesystem.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_COMPRESSION 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_COMPRESSION, VER_KERNEL,
   "Rocklyte Systems, Jean-loup Gailly, Mark Adler",
   "Rocklyte Systems (c) 2001-2004.  All rights reserved.",
   "April 2004",
   "Compression"
};

static struct KernelBase *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static struct FileSystemBase *FileSystemBase = 0;
static OBJECTPTR modStrings = 0, modFileSystem = 0;
static OBJECTPTR CompressionClass = 0;

/*****************************************************************************
** Class definition.
*/

#define LEN_PASSWORD 128
#define COMPRESS_LEN_HEADER 32

struct ZipFile {
   FIELDS_COMPRESSEDFILE

   /*** Private fields ***/
   ULONG TimeStamp;              /* Time stamp information */
   ULONG CRC;                    /* CRC validation number */
   ULONG Offset;                 /* Byte offset of the file within the archive */
   WORD  NameLen;                /* Length of name string */
   WORD  CommentLen;             /* Length of comment string */
   WORD  DeflateMethod;          /* Set to 8 for normal deflation */
   LONG  Flags;                  /* System specific flags for the file entry */
};

struct Compression {
   OBJECT_HEADER
   OBJECTID OutputID;                    /* Reference to output object for user messages */
   LONG     CompressionLevel;            /* Compression level (percentage - 0% none, 100% high) */
   LONG     Flags;                       /* Special flags */
   LONG     SegmentSize;                 /* Splits the compressed file if it surpasses a set byte limit (e.g. 1.44MB) */
   LONG     Static;                      /* Set if the object should reside in memory */
   STRING   Location;                    /* File location of the compressed data */
   STRING   Destination;                 /* Destination path for decompressing data */
   STRING   *FileList;                   /* List of all files held in the compression object */
   LONG     (*Feedback)(struct Compression *, struct CompressionFeedback *);
   LONG     Permissions;                 /* Default permissions to use for decompressed files */

   PRIVATE_FIELDS
   UBYTE    Header[COMPRESS_LEN_HEADER]; /* The first 32 bytes of data from the compressed file (for sub-classes only) */
   UBYTE    Password[LEN_PASSWORD];      /* Password for the compressed object */
   OBJECTPTR FileIO;                     /* File input/output */

   /*** Zip only fields ***/
   WORD   prvCompressionCount;             /* Counter of times that compression has occurred */
   z_stream prvZip;
   UBYTE  *prvOutput;
   UBYTE  *prvInput;
   struct ZipFile *prvFiles;           /* List of files in the archive (must be in order of the archive's entries) */
   LONG   prvTotalFiles;
   LONG   prvFileIndex;
};

static ERROR GET_FileList(struct Compression *, struct CompressedFile **);
static ERROR GET_Header(struct Compression *, UBYTE **);
static ERROR GET_Location(struct Compression *, STRING *);
static ERROR GET_Password(struct Compression *, STRING *);
static ERROR GET_Size(struct Compression *, LARGE *);
static ERROR GET_UncompressedSize(struct Compression *, LARGE *);

static ERROR SET_CompressionLevel(struct Compression *, LONG *);
static ERROR SET_Location(struct Compression *, STRING);
static ERROR SET_Password(struct Compression *, STRING);

struct FieldDef CompressionFlags[] = {
   { "PASSWORD",      CMF_PASSWORD },
   { "NEW",           CMF_NEW },
   { "CREATEFILE",    CMF_CREATEFILE },
   { "READONLY",      CMF_READONLY },
   { "NOLINKS",       CMF_NOLINKS },
   { "APPLYSECURITY", CMF_APPLYSECURITY },
   { NULL, NULL }
};

static struct FieldDef PermissionFlags[] = {
   { "READ",          PERMIT_READ },
   { "WRITE",         PERMIT_WRITE },
   { "EXEC",          PERMIT_EXEC },
   { "EXECUTABLE",    PERMIT_EXEC },
   { "DELETE",        PERMIT_DELETE },
   { "HIDDEN",        PERMIT_HIDDEN },
   { "ARCHIVE",       PERMIT_ARCHIVE },
   { "PASSWORD",      PERMIT_PASSWORD },
   { "USERID",        PERMIT_USERID },
   { "GROUPID",       PERMIT_GROUPID },
   { "OTHERS_READ",   PERMIT_OTHERS_READ },
   { "OTHERS_WRITE",  PERMIT_OTHERS_WRITE },
   { "OTHERS_EXEC",   PERMIT_OTHERS_EXEC },
   { "OTHERS_DELETE", PERMIT_OTHERS_DELETE },
   { "GROUP_READ",    PERMIT_GROUP_READ },
   { "GROUP_WRITE",   PERMIT_GROUP_WRITE },
   { "GROUP_EXEC",    PERMIT_GROUP_EXEC },
   { "GROUP_DELETE",  PERMIT_GROUP_DELETE },
   { "ALL_READ",      PERMIT_ALL_READ },
   { "ALL_WRITE",     PERMIT_ALL_WRITE },
   { "ALL_EXEC",      PERMIT_ALL_EXEC },
   { NULL, NULL }
};

static struct FieldArray CompressionFields[] = {
   { "Output",           0, FDF_OBJECTID|FDF_RI,  0, NULL, NULL },
   { "CompressionLevel", 0, FDF_LONG|FDF_RW,      0, NULL, SET_CompressionLevel },
   { "Flags",            0, FDF_LONGFLAGS|FDF_RW, (LONG)&CompressionFlags, NULL, NULL },
   { "SegmentSize",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Static",           0, FDF_LONG|FDF_RI,      0, NULL, NULL },
   { "Location",         0, FDF_STRING|FDF_RW,    0, NULL, SET_Location },
   { "Destination",      0, FDF_STRING|FDF_RW,    0, NULL, NULL },
   { "FileList",         0, FDF_POINTER|FDF_R,    0, GET_FileList, NULL },
   { "Feedback",         0, FDF_POINTER|FDF_RW,   0, NULL, NULL },
   { "Permissions",      0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&PermissionFlags, NULL, NULL },
   /*** Virtual fields ***/
   { "Header",           0, FDF_POINTER|FDF_R,   0, GET_Header,   NULL },
   { "Password",         0, FDF_STRING|FDF_RW,   0, GET_Password, SET_Password },
   { "Size",             0, FDF_LARGE|FDF_R,     0, GET_Size, NULL },
   { "Src",              0, FDF_STRING|FDF_RW,   0, GET_Location, SET_Location },
   { "UncompressedSize", 0, FDF_LARGE|FDF_R,     0, GET_UncompressedSize, NULL },
   END_FIELD
};

/*****************************************************************************
** Compression actions.
*/

static ERROR COMPRESS_ClosingTag(struct Compression *, APTR);
static ERROR COMPRESS_Flush(struct Compression *, APTR);
static ERROR COMPRESS_Free(struct Compression *, APTR);
static ERROR COMPRESS_HealthCheck(struct Compression *, APTR);
static ERROR COMPRESS_Init(struct Compression *, APTR);
static ERROR COMPRESS_NewObject(struct Compression *, APTR);

static struct ActionArray CompressionActions[] = {
   { AC_ClosingTag,  COMPRESS_ClosingTag },
   { AC_Flush,       COMPRESS_Flush },
   { AC_Free,        COMPRESS_Free },
   { AC_HealthCheck, COMPRESS_HealthCheck },
   { AC_Init,        COMPRESS_Init },
   { AC_NewObject,   COMPRESS_NewObject },
   { NULL, NULL }
};

/*****************************************************************************
** Compression methods.
*/

static ERROR COMPRESS_CompressBuffer(struct Compression *, struct mtCompressBuffer *);
static ERROR COMPRESS_CompressFile(struct Compression *, struct mtCompressFile *);
static ERROR COMPRESS_DecompressBuffer(struct Compression *, struct mtDecompressBuffer *);
static ERROR COMPRESS_DecompressFile(struct Compression *, struct mtDecompressFile *);
static ERROR COMPRESS_RemoveFile(struct Compression *, struct mtRemoveFile *);

static struct FunctionField argsCompressBuffer[]   = { { NULL, NULL } };
static struct FunctionField argsCompressFile[]     = { { "Location", ARG_STRING }, { "Path", ARG_STRING }, { NULL, NULL } };
static struct FunctionField argsDecompressBuffer[] = { { NULL, NULL } };
static struct FunctionField argsDecompressFile[]   = { { "Path", ARG_STRING }, { "Destination", ARG_STRING }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsRemoveFile[]       = { { "Path", ARG_STRING }, { NULL, NULL } };

static struct MethodArray CompressionMethods[] = {
   { MT_CompressFile,     COMPRESS_CompressFile,     "CompressFile",     argsCompressFile,     sizeof(struct mtCompressFile) },
   { MT_CompressBuffer,   COMPRESS_CompressBuffer,   "CompressBuffer",   argsCompressBuffer,   sizeof(struct mtCompressBuffer) },
   { MT_DecompressBuffer, COMPRESS_DecompressBuffer, "DecompressBuffer", argsDecompressBuffer, sizeof(struct mtDecompressBuffer) },
   { MT_DecompressFile,   COMPRESS_DecompressFile,   "DecompressFile",   argsDecompressFile,   sizeof(struct mtDecompressFile) },
   { MT_RemoveFile,       COMPRESS_RemoveFile,       "RemoveFile",       argsRemoveFile,       sizeof(struct mtRemoveFile) },
   { NULL, NULL }
};

/*****************************************************************************
** Special definitions.
*/

#define SIZE_COMPRESSION_BUFFER 16384

#define ZIP_ATHENE 0x7e /* Use this identifier to declare Athene zipped files */

/* The following flags can be tagged to each file entry in the zip file and
** are Athene-specific (identifiable by the ZIP_ATHENE OS tag).
**
** NOTE: Originally we used the low order bits, this was a mistake because
** WinZip, WinRar and so forth assume that those bits have meaning.  So
** our flags have ended up being doubled up for the moment so that we can
** continue to read .install packages that were compiled with the old flags.
*/

#define ZIP_LINK   0x00010001 /* The entry is a symbolic link */
#define ZIP_UEXEC  0x00020002 /* Executable-access allowed (owner) */
#define ZIP_GEXEC  0x00040004 /* Executable-access allowed (group) */
#define ZIP_OEXEC  0x00080008 /* Executable-access allowed (everyone) */
#define ZIP_UREAD  0x00100010 /* Read-access allowed (owner) */
#define ZIP_GREAD  0x00200020 /* Read-access allowed (group) */
#define ZIP_OREAD  0x00400040 /* Read-access allowed (everyone) */
#define ZIP_UWRITE 0x00800080 /* Write-access allowed (owner) */
#define ZIP_GWRITE 0x01000100 /* Write-access allowed (group) */
#define ZIP_OWRITE 0x02000200 /* Write-access allowed (everyone) */

#define ZIP_SECURITY (ZIP_UEXEC | ZIP_GEXEC | ZIP_OEXEC | ZIP_UREAD | ZIP_GREAD | ZIP_OREAD | ZIP_UWRITE | ZIP_GWRITE | ZIP_OWRITE)

#define HEAD_DEFLATEMETHOD  8
#define HEAD_TIMESTAMP      10
#define HEAD_CRC            14
#define HEAD_COMPRESSEDSIZE 18
#define HEAD_FILESIZE       22
#define HEAD_NAMELEN        26   /* File name */
#define HEAD_EXTRALEN       28   /* System specific information */
#define HEAD_LENGTH         30

#define LIST_SIGNATURE      0
#define LIST_VERSION        4
#define LIST_OS             5
#define LIST_REQUIRED_VER   6
#define LIST_REQUIRED_OS    7
#define LIST_FLAGS          8
#define LIST_METHOD         10
#define LIST_TIMESTAMP      12
#define LIST_CRC            16  /* Checksum */
#define LIST_COMPRESSEDSIZE 20
#define LIST_FILESIZE       24  /* Original file size */
#define LIST_NAMELEN        28  /* File name */
#define LIST_EXTRALEN       30  /* System specific information */
#define LIST_COMMENTLEN     32  /* Optional comment */
#define LIST_DISKNO         34  /* Disk number start */
#define LIST_IFILE          36  /* Internal file attributes (pkzip specific) */
#define LIST_ATTRIB         38  /* System specific file attributes */
#define LIST_OFFSET         42  /* Relative offset of local header */
#define LIST_LENGTH         46

#define TAIL_FILECOUNT1     8
#define TAIL_FILECOUNT2     10
#define TAIL_FILELISTSIZE   12
#define TAIL_FILELISTOFFSET 16
#define TAIL_LENGTH         22

static UBYTE header[HEAD_LENGTH] = {
   'P', 'K', 0x03, 0x04,   /* 0 Signature */
   0x14, 0x00,             /* 4 Version 2.0 */
   0x00, 0x00,             /* 6 Flags */
   0x08, 0x00,             /* 8 Deflation method */
   0x00, 0x00, 0x00, 0x00, /* 10 Time stamp */
   0x00, 0x00, 0x00, 0x00, /* 14 CRC */
   0x00, 0x00, 0x00, 0x00, /* 18 Compressed Size */
   0x00, 0x00, 0x00, 0x00, /* 22 Original File Size */
   0x00, 0x00,             /* 26 Length of path/filename */
   0x00, 0x00              /* 28 Length of extra field. */
};

static UBYTE listdata[LIST_LENGTH] = {
   'P', 'K', 0x01, 0x02,   /* 00 Signature */
   0x14, ZIP_ATHENE,       /* 04 Version 2.0, host OS */
   0x14, 0x00,             /* 06 Version need to extract, OS needed to extract */
   0x00, 0x00,             /* 08 Flags */
   0x08, 0x00,             /* 10 Deflation method */
   0x00, 0x00, 0x00, 0x00, /* 12 Time stamp */
   0x00, 0x00, 0x00, 0x00, /* 16 CRC */
   0x00, 0x00, 0x00, 0x00, /* 20 Compressed Size */
   0x00, 0x00, 0x00, 0x00, /* 24 Original File Size */
   0x00, 0x00,             /* 28 Length of path/filename */
   0x00, 0x00,             /* 30 Length of extra field */
   0x00, 0x00,             /* 32 Length of comment */
   0x00, 0x00,             /* 34 Disk number start */
   0x00, 0x00,             /* 36 File attribute: 0 = Binary, 1 = ASCII */
   0x00, 0x00, 0x00, 0x00, /* 38 File permissions? */
   0x00, 0x00, 0x00, 0x00, /* 42 Offset of compressed data within the file */
   /*** File name follows ***/
   /*** Extra field follows ***/
   /*** Comment follows ***/
};

static UBYTE taildata[TAIL_LENGTH] = {
   'P', 'K', 0x05, 0x06,   /* 0 Signature */
   0x00, 0x00,             /* 4 Number of this disk */
   0x00, 0x00,             /* 6 Number of the disk with starting central directory */
   0x00, 0x00,             /* 8 Number of files in the central directory for this zip file */
   0x00, 0x00,             /* 10 Number of files in the archive spanning all disks */
   0x00, 0x00, 0x00, 0x00, /* 12 Size of file list */
   0x00, 0x00, 0x00, 0x00, /* 16 Offset of the file list with respect to starting disk number */
   0x00, 0x00              /* 20 Length of zip file comment */
   /*** End of file comment follows ***/
};

/*****************************************************************************
** Stubs.
*/

static LONG ReadLong(OBJECTPTR File)
{
   struct acRead args;
   LONG value;
   args.Buffer = (APTR)&value;
   args.Length = 4;
   if (Action(AC_Read, File, &args) IS ERR_Okay) return(value); /* !!! Use ReadLE/BE */
   else DPrintF("@ReadLong()","Failed.");
   return(0);
}

static WORD ReadWord(OBJECTPTR File)
{
   struct acRead args;
   WORD value;
   args.Buffer = (APTR)&value;
   args.Length = 2;
   if (Action(AC_Read, File, &args) IS ERR_Okay) return(value); /* !!! Use ReadLE/BE */
   else DPrintF("@ReadWord()","Failed.");
   return(0);
}

static LONG Write(OBJECTPTR File, APTR Buffer, LONG Size)
{
   struct acWrite write;
   write.Buffer = Buffer;
   write.Length = Size;
   if (Action(AC_Write, File, &write) IS ERR_Okay) return(write.Result);
   else return(NULL);
}

/*****************************************************************************
** Function prototypes.
*/

static ERROR CompressDirectory(struct Compression *, STRING, STRING);
static ERROR CompressFile(struct Compression *, STRING, STRING, BYTE);
static void Print(struct Compression *, STRING);
static ERROR RemoveFile(struct Compression *, struct ZipFile **);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&CompressionClass, NULL,
      FID_BaseClassID|TLONG, ID_COMPRESSION,
      FID_Version|TFLOAT,    VER_COMPRESSION,
      FID_Name|TSTRING,      "Compression",
      FID_Category|TLONG,    CCF_DATA,
      FID_Actions|TPTR,      CompressionActions,
      FID_Methods|TPTR,      CompressionMethods,
      FID_Fields|TPTR,       CompressionFields,
      FID_Size|TLONG,        sizeof(struct Compression),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (CompressionClass) { acFree(CompressionClass); CompressionClass = NULL; }
   if (modStrings)       { acFree(modStrings);       modStrings = NULL; }
   if (modFileSystem)    { acFree(modFileSystem);    modFileSystem = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  ClosingTag
Short: If the object is non-static, this action will start the execution process and then the object will self-destruct.
###END###

*****************************************************************************/

static ERROR COMPRESS_ClosingTag(struct Compression *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      acActivate(Self);
      acFree(Self);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  CompressBuffer
Short: Compresses a plain memory area into an empty buffer.

###DESCRIPTION###
<p>This method provides a simple way of compressing a memory area into a
buffer.  You need to provide a reference to the source data and a buffer
large enough to accept the compressed information.  Generally your
destination buffer should be no smaller than 75% of the size of the
source data.  If the destination buffer is not large enough, an error of
ERR_BufferOverflow will be returned.  The size of the compressed data
will be returned in the Result parameter.</p>

<p>You can decompress the data that is output by this function by using the
<a href="#DecompressBuffer()">DecompressBuffer</a> method.</p>

###INPUT###
APTR/Input:      Pointer to the source data.
LONG/InputSize:  Byte length of the source data.
APTR/Output:     Pointer to a destination buffer.
LONG/OutputSize: Available space in the destination buffer.
LONG/Result:     The size of the compressed data will be returned in this parameter.

###RESULT###
ERR_Okay: The data was compressed successfully.  The Result parameter indicates the size of the compressed data.
ERR_Args: Invalid arguments were detected.
ERR_BufferOverflow: The output buffer is not large enough.

###SEE ALSO###
Method: DecompressBuffer

###END###

*****************************************************************************/

static ERROR COMPRESS_CompressBuffer(struct Compression *Self, struct mtCompressBuffer *Args)
{
   LONG err, level;

   if ((!Args) OR (!Args->Input) OR (Args->InputSize <= 0) OR (!Args->Output) OR (Args->OutputSize <= 8)) {
      return(ObjectError(Self, MT_CompressBuffer, ERR_Args));
   }

   Self->prvZip.next_in   = Args->Input;
   Self->prvZip.avail_in  = Args->InputSize;
   Self->prvZip.next_out  = Args->Output + 8;
   Self->prvZip.avail_out = Args->OutputSize - 8;

   level = Self->CompressionLevel / 10;
   if (level < 0) level = 0;
   else if (level > 9) level = 9;

   if ((err = deflateInit2(&Self->prvZip, level, Z_DEFLATED, -MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY)) IS ERR_Okay) {
      if ((err = deflate(&Self->prvZip, Z_FINISH)) IS Z_STREAM_END) {
         Args->Result = Self->prvZip.total_out;
         deflateEnd(&Self->prvZip);

         ((BYTE *)Args->Output)[0] = 'Z';
         ((BYTE *)Args->Output)[1] = 'L';
         ((BYTE *)Args->Output)[2] = 'I';
         ((BYTE *)Args->Output)[3] = 'B';
         ((LONG *)Args->Output)[1] = Args->Result;
         return(ERR_Okay);
      }
      else {
         deflateEnd(&Self->prvZip);
         return(ObjectError(Self, MT_CompressBuffer, ERR_BufferOverflow));
      }
   }
   else return(ObjectError(Self, MT_CompressBuffer, ERR_Failed));
}

/*****************************************************************************

###METHOD###
Name:  CompressFile
Short: Add files to a compression object.

###DESCRIPTION###
<p>The CompressFile method is used to add new files and folders to a compression object.
You need to supply the location of the file that you wish to add, as well
as the path that you want the file to use when it is in the compression
object.  The Location parameter accepts wildcards, which gives you the ability
to add multiple files in a single function call if you require this
convenience.  If you want to add an entire directory, specify its complete
path in the Location parameter and ensure that it ends with a trailing
slash or assignment character.  The Path parameter must end with a forward
slash in all cases in order to indicate a directory name, unless your intention
is to archive a single file using a name that is different to its own.</p>

###INPUT###
STRING/Location: The location of the file(s) to add.
STRING/Path:     The path that you want to use for this file when it is added to the compression object.  May be NULL for no path.

###RESULT###
ERR_Okay: The file was added to the compression object.
ERR_Args: Invalid arguments were detected.
ERR_File: An error was encountered while trying to open the source file.
ERR_NoPermission: The CMF_READONLY flag has been set on the compression object.

###END###

*****************************************************************************/

static ERROR COMPRESS_CompressFile(struct Compression *Self, struct mtCompressFile *Args)
{
   STRING path;
   UBYTE wildcard;
   LONG i, j, len, pathlen;
   ERROR error;

   if ((!Args) OR (!Args->Location) OR (!*Args->Location)) return(ObjectError(Self, MT_CompressFile, ERR_Args));
   if (!Self->FileIO) return(ObjectError(Self, MT_CompressFile, ERR_MissingLocation));

   if (Self->Flags & CMF_READONLY) return(ObjectError(Self, MT_CompressFile, ERR_NoPermission));

   if (Self->Head.SubID) return(ObjectError(Self, MT_CompressFile, ERR_NoSupport));

   if (Self->OutputID) {
      StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "Compressing \"%s\" to \"%s\".\n", Args->Location, Self->Location);
      Print(Self, Self->prvOutput);
   }

   if (!Args->Path) path = "";
   else {
      /*** Accept the path by default but check it for illegal symbols just in case ***/

      path = Args->Path;
      for (i=0; path[i]; i++) {
         if ((path[i] IS '*') OR (path[i] IS '?') OR (path[i] IS '"') OR
             (path[i] IS ':') OR (path[i] IS '|') OR (path[i] IS '<') OR
             (path[i] IS '>')) {
            DPrintF("@CompressFile:","Illegal characters in path: %s", path);
            if (Self->OutputID) {
               StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "Warning - path ignored due to illegal characters: %s\n", path);
               Print(Self, Self->prvOutput);
            }
            path = "";
            break;
         }
      }
   }

   DPrintF("CompressFile()","Location: %s, Path: %s", Args->Location, path);

   Self->prvFileIndex = 0;

   for (i=0; Args->Location[i]; i++);
   if ((Args->Location[i-1] IS '/') OR (Args->Location[i-1] IS '\\') OR (Args->Location[i-1] IS ':')) {
      if (*path) {
         /* The path must be an empty string or terminated
         ** with a slash character.
         */
         for (i=0; path[i]; i++);
         if (path[i-1] != '/') {
            UBYTE newpath[i+2];
            for (i=0; path[i]; i++) newpath[i] = path[i];
            newpath[i++] = '/';
            newpath[i] = 0;
            return(CompressDirectory(Self, Args->Location, newpath));
         }
      }

      return(CompressDirectory(Self, Args->Location, path));
   }

   error = ERR_Okay;

   /*** Check the location string for wildcards, * and ? ***/

   wildcard = FALSE;
   for (len=0; Args->Location[len]; len++);
   for (pathlen=len; pathlen > 0; pathlen--) {
      if ((Args->Location[pathlen-1] IS '*') OR (Args->Location[pathlen-1] IS '?')) wildcard = TRUE;
      else if ((Args->Location[pathlen-1] IS ':') OR (Args->Location[pathlen-1] IS '/') OR (Args->Location[pathlen-1] IS '\\')) break;
   }

   if (wildcard IS FALSE) return(CompressFile(Self, Args->Location, path, FALSE));
   else {
      struct FileInfo *list, *scan;
      UBYTE filename[len-pathlen+1];
      UBYTE dirlocation[len+1];

      /*** Extract the file name from the location (drop the path) ***/

      for (j=0; Args->Location[pathlen+j]; j++) filename[j] = Args->Location[pathlen+j];
      filename[j] = 0;

      /*** Extract the path from the location (drop the file name) ***/

      for (j=0; j < pathlen; j++) dirlocation[j] = Args->Location[j];
      dirlocation[j] = 0;

      if (ReadDirectory(dirlocation, RDF_FILE, &list) IS ERR_Okay) {
         for (scan=list; (scan) AND (error IS ERR_Okay); scan=scan->Next) {
            if (StrCompare(filename, scan->Name, NULL, STR_WILDCARD) IS ERR_Okay) {
               for (len=0; scan->Name[len]; len++);
               UBYTE directory[pathlen+len+1];
               for (j=0; j < pathlen; j++) directory[j] = Args->Location[j];
               for (j=0; scan->Name[j]; j++) directory[pathlen+j] = scan->Name[j];
               directory[pathlen+j] = 0;
               error = CompressFile(Self, directory, path, FALSE);
            }
         }

         FreeMemory(list);
      }
   }

   if (Self->OutputID) {
      LARGE size;
      GET_Size(Self, &size);
      StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "\nCompression complete.  Archive is %d bytes in size.", size);
      Print(Self, Self->prvOutput);
   }

   return(error);
}

/*****************************************************************************

###METHOD###
Name:  DecompressBuffer
Short: Decompresses data originating from the CompressBuffer method.

###DESCRIPTION###
<p>This method is used to decompress data that has been packed using the
<a href="#CompressBuffer()">CompressBuffer</a> method.  You need to provide
a pointer to the compressed information and an output buffer large enough to
accept the decompressed information.  If the output buffer is not large enough
to contain the data, the method will write out as much information as it can
and then return with an error code of ERR_BufferOverflow.</p>

###INPUT###
APTR/Input:      Pointer to the compressed data.
APTR/Output:     Pointer to the decompression buffer.
LONG/OutputSize: Size of the decompression buffer.
LONG/Result:     The amount of bytes decompressed will be returned in this parameter.

###RESULT###
ERR_Okay: The data was decompressed successfully.
ERR_Args: Invalid arguments were detected.
ERR_BufferOverflow: The output buffer is not large enough to hold the decompressed information.

###END###

*****************************************************************************/

static ERROR COMPRESS_DecompressBuffer(struct Compression *Self, struct mtDecompressBuffer *Args)
{
   LONG err;

   if ((!Args) OR (!Args->Input) OR (!Args->Output) OR (Args->OutputSize <= 0)) {
      return(ObjectError(Self, MT_DecompressBuffer, ERR_Args));
   }

   Self->prvZip.next_in   = Args->Input + 8;
   Self->prvZip.avail_in  = ((LONG *)Args->Input)[1];
   Self->prvZip.next_out  = Args->Output;
   Self->prvZip.avail_out = Args->OutputSize;

   if (inflateInit2(&Self->prvZip, -MAX_WBITS) IS ERR_Okay) {
      if ((err = inflate(&Self->prvZip, Z_FINISH)) IS Z_STREAM_END) {
         Args->Result = Self->prvZip.total_out;
         inflateEnd(&Self->prvZip);
         return(ERR_Okay);
      }
      else {
         inflateEnd(&Self->prvZip);
         return(ObjectError(Self, MT_DecompressBuffer, ERR_BufferOverflow));
      }
   }
   else return(ObjectError(Self, MT_DecompressBuffer, ERR_Failed));
}

/*****************************************************************************

###METHOD###
Name:  DecompressFile
Short: Extracts one or more files from a compression object.

###DESCRIPTION###
<p>Use the DecompressFile method to decompress a file or files to a destination
directory.  You need to know the exact path name of the file that you want
to extract, unless you wish to use wildcards.  A single asterisk in the Path
parameter will extract all files in a compression object.</p>

<p>When specifing the Destination parameter, it is recommended that you
specify a directory location by using a forward slash at the end of the
string.  Otherwise the Destination will be interpreted as a specific file
name.  If the destination file already exists, it will be overwritten
by the decompressed data.</p>

###INPUT###
STRING/Path: The full path name of the file to extract from the archive.
STRING/Destination: The destination to extract the file to.
LONG/Flags: Optional flags.

###RESULT###
ERR_Okay: The file was successfully extracted.
ERR_Args: Invalid arguments were detected.
ERR_File: A destination file could not be created.
ERR_Seek: A seek error occurred
ERR_Write: Failed to write uncompressed information to a destination file.
ERR_Cancelled: The decompression process was cancelled by the feedback mechanism.
ERR_Failed: General decompression error.

###END###

*****************************************************************************/

static ERROR COMPRESS_DecompressFile(struct Compression *Self, struct mtDecompressFile *Args)
{
   struct acRead read;
   struct acWrite write;
   struct CompressionFeedback feedback;
   struct ZipFile *filelist;
   UBYTE location[400];
   OBJECTPTR file;
   UWORD i, j, pos, inflateend, pathend, namelen, extralen;
   LONG err, inputlen, result;
   ERROR error;

   /*** Validate arguments ***/

   if ((!Args) OR (!Args->Path)) {
      if (Self->OutputID) {
         StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "Please supply a Path setting that refers to a compressed file archive.\n");
         Print(Self, Self->prvOutput);
      }

      return(ObjectError(Self, MT_DecompressFile, ERR_Args));
   }

   if (!Args->Destination) {
      if (Self->OutputID) {
         StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "Please supply a Destination that refers to a directory for decompression.\n");
         Print(Self, Self->prvOutput);
      }

      return(ObjectError(Self, MT_DecompressFile, ERR_Args));
   }

   if ((!*Args->Path) OR (!*Args->Destination)) {
      if (Self->OutputID) {
         StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "Please supply valid Path and Destination settings.\n");
         Print(Self, Self->prvOutput);
      }

      return(ObjectError(Self, MT_DecompressFile, ERR_Args));
   }

   if (!Self->FileIO) {
      if (Self->OutputID) {
         StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "Internal error - decompression aborted.\n");
         Print(Self, Self->prvOutput);
      }

      return(ObjectError(Self, MT_DecompressFile, ERR_MissingLocation));
   }

   /*** If the object belongs to a Compression sub-class, return ERR_NoSupport ***/

   if (Self->Head.SubID) return(ERR_NoSupport);

   /*** Tell the user what we are doing ***/

   if (Self->OutputID) {
      StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "Decompressing archive \"%s\" with path \"%s\" to \"%s\".\n", Self->Location, Args->Path, Args->Destination);
      Print(Self, Self->prvOutput);
   }

   /* Search for the file(s) in our archive that match the given name
   ** and extract them to the destination directory.
   */

   DPrintF("~DecompressFile()","%s TO %s, Permissions: $%.8x", Args->Path, Args->Destination, Self->Permissions);

   for (pos=0; (Args->Destination[pos]) AND (pos < sizeof(location)); pos++) location[pos] = Args->Destination[pos];
   location[pos] = 0;

   pathend = 0;
   for (i=0; Args->Path[i]; i++) if (Args->Path[i] IS '/') pathend = i + 1;

   error      = ERR_Okay;
   inflateend = FALSE;
   file       = NULL;
   Self->prvFileIndex = 0;

   for (filelist=Self->prvFiles; (filelist); filelist=(struct ZipFile *)filelist->Next) {
      if (StrCompare(Args->Path, filelist->Name, 0, STR_WILDCARD) IS ERR_Okay) {
         DPrintF("DecompressFile:","[Compression:%d] Extracting \"%s\"", Self->Head.UniqueID, filelist->Name);

         if (Self->OutputID) {
            StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "  %s", filelist->Name);
            Print(Self, Self->prvOutput);
         }

         /* If the destination path specifies a directory, add the name of the
         ** file to the destination to generate the correct file name.
         */

         j = pos;
         if ((location[j-1] IS '/') OR (location[j-1] IS '\\') OR (location[j-1] IS ':')) {
            for (i=pathend; filelist->Name[i]; i++) {
               location[j++] = filelist->Name[i];
            }
            location[j] = 0;
         }

         /*** If the destination is a directory that already exists, skip this compression entry ***/

         if ((location[j-1] IS '/') OR (location[j-1] IS '\\')) {
            if ((AnalyseLocation(location, &result) IS ERR_Okay) AND (result IS LOC_DIRECTORY)) {
               Self->prvFileIndex++;
               continue;
            }
         }

         /*** Send compression feedback if requested ***/

         feedback.Year   = 1980 + ((filelist->TimeStamp>>25) & 0x3f);
         feedback.Month  = (filelist->TimeStamp>>21) & 0x0f;
         feedback.Day    = (filelist->TimeStamp>>16) & 0x1f;
         feedback.Hour   = (filelist->TimeStamp>>11) & 0x1f;
         feedback.Minute = (filelist->TimeStamp>>5)  & 0x3f;
         feedback.Second = (filelist->TimeStamp>>1)  & 0x0f;

         if (Self->Feedback) {
            feedback.FeedbackID     = FDB_DecompressFile;
            feedback.Index          = Self->prvFileIndex;
            feedback.Path           = filelist->Name;
            feedback.Destination    = location;
            feedback.OriginalSize   = filelist->OriginalSize;
            feedback.CompressedSize = filelist->CompressedSize;
            feedback.Progress       = 0;
            ClearMemory(&feedback.Reserved, sizeof(feedback.Reserved));

            result = Self->Feedback(Self, &feedback);
            if (result IS CFR_Abort) {
               error = ObjectError(Self, MT_DecompressFile, ERR_Cancelled);
               goto exit;
            }
            else if (result IS CFR_Skip) {
               Self->prvFileIndex++; /* Increase counter to show that the file was analysed */
               continue;
            }
         }

         /*** Seek to the start of the compressed data ***/

         if (ActionTags(AC_Seek, Self->FileIO, (DOUBLE)(filelist->Offset + HEAD_NAMELEN), SEEK_START) != ERR_Okay) {
            error = ObjectError(Self, MT_DecompressFile, ERR_Seek);
            goto exit;
         }

         namelen = ReadWord(Self->FileIO);
         extralen = ReadWord(Self->FileIO);
         if (ActionTags(AC_Seek, Self->FileIO, (DOUBLE)(namelen + extralen), SEEK_CURRENT) != ERR_Okay) {
            error = ObjectError(Self, MT_DecompressFile, ERR_Seek);
            goto exit;
         }

         if (filelist->Flags & ZIP_LINK) {
            /* For symbolic links, decompress the data to get the destination
            ** link string
            */

            Self->prvZip.next_in   = 0;
            Self->prvZip.avail_in  = 0;
            Self->prvZip.next_out  = 0;
            Self->prvZip.avail_out = 0;

            if (filelist->CompressedSize > 0) {
               if (filelist->DeflateMethod IS 0) {
                  /*** This routine is used if the file is stored rather than compressed ***/

                  inputlen = filelist->CompressedSize;

                  read.Buffer = Self->prvInput;
                  read.Length = SIZE_COMPRESSION_BUFFER-1;
                  if ((error = Action(AC_Read, Self->FileIO, &read)) IS ERR_Okay) {
                     Self->prvInput[read.Result] = 0;
                     DeleteFile(location);
                     error = CreateLink(location, Self->prvInput);
                     if (error IS ERR_NoSupport) error = ERR_Okay;
                  }

                  if (error) goto exit;
               }
               else if ((filelist->DeflateMethod IS 8) AND (inflateInit2(&Self->prvZip, -MAX_WBITS) IS ERR_Okay)) {
                  inflateend = TRUE;

                  read.Buffer = Self->prvInput;
                  if (filelist->CompressedSize < SIZE_COMPRESSION_BUFFER) read.Length = filelist->CompressedSize;
                  else read.Length = SIZE_COMPRESSION_BUFFER;

                  err = ERR_Okay;
                  if ((error = Action(AC_Read, Self->FileIO, &read)) != ERR_Okay) goto exit;
                  if (read.Result <= 0) { error = ERR_Read; goto exit; }
                  inputlen = filelist->CompressedSize - read.Result;

                  Self->prvZip.next_in   = Self->prvInput;
                  Self->prvZip.avail_in  = read.Result;
                  Self->prvZip.next_out  = Self->prvOutput;
                  Self->prvZip.avail_out = SIZE_COMPRESSION_BUFFER-1;

                  err = inflate(&Self->prvZip, Z_SYNC_FLUSH);

                  if ((err != ERR_Okay) AND (err != Z_STREAM_END)) {
                     if (Self->prvZip.msg) DPrintF("!DecompressFile:","%s", Self->prvZip.msg);
                     error = ERR_Failed;
                     goto exit;
                  }

                  Self->prvOutput[filelist->OriginalSize] = 0; /* !!! We should terminate according to the amount of data decompressed */
                  DeleteFile(location);
                  error = CreateLink(location, Self->prvOutput);
                  if (error IS ERR_NoSupport) error = ERR_Okay;

                  inflateEnd(&Self->prvZip);
                  inflateend = FALSE;
               }
            }
         }
         else {
            /*** Create the destination file or directory ***/

            LONG permissions;

            if (Self->Flags & CMF_APPLYSECURITY) {
               if (filelist->Flags & ZIP_SECURITY) {
                  permissions = NULL;
                  if (filelist->Flags & ZIP_UEXEC) permissions |= PERMIT_USER_EXEC;
                  if (filelist->Flags & ZIP_GEXEC) permissions |= PERMIT_GROUP_EXEC;
                  if (filelist->Flags & ZIP_OEXEC) permissions |= PERMIT_OTHERS_EXEC;

                  if (filelist->Flags & ZIP_UREAD) permissions |= PERMIT_USER_READ;
                  if (filelist->Flags & ZIP_GREAD) permissions |= PERMIT_GROUP_READ;
                  if (filelist->Flags & ZIP_OREAD) permissions |= PERMIT_OTHERS_READ;

                  if (filelist->Flags & ZIP_UWRITE) permissions |= PERMIT_USER_WRITE;
                  if (filelist->Flags & ZIP_GWRITE) permissions |= PERMIT_GROUP_WRITE;
                  if (filelist->Flags & ZIP_OWRITE) permissions |= PERMIT_OTHERS_WRITE;
               }
               else permissions = Self->Permissions;
            }
            else permissions = Self->Permissions;

            if ((error = CreateObject(ID_FILE, NF_CHILD, &file, NULL,
               FID_Location|TSTRING,  location,
               FID_Flags|TLONG,       FL_NEW|FL_WRITE,
               FID_Permissions|TLONG, permissions,
               TAGEND)) != ERR_Okay) {
               DPrintF("@DecompressFile:","Error %d creating file \"%s\".", error, location);
               goto exit;
            }

            Self->prvZip.next_in   = 0;
            Self->prvZip.avail_in  = 0;
            Self->prvZip.next_out  = 0;
            Self->prvZip.avail_out = 0;

            if (filelist->CompressedSize > 0) {
               if (filelist->DeflateMethod IS 0) {
                  /*** This routine is used if the file is stored rather than compressed ***/

                  inputlen = filelist->CompressedSize;

                  read.Buffer = Self->prvInput;
                  if (inputlen < SIZE_COMPRESSION_BUFFER) read.Length = inputlen;
                  else read.Length = SIZE_COMPRESSION_BUFFER;

                  while (((error = Action(AC_Read, Self->FileIO, &read)) IS ERR_Okay) AND (read.Result > 0)) {
                     write.Buffer = Self->prvInput;
                     write.Length = read.Result;
                     if (Action(AC_Write, file, &write) != ERR_Okay) { error = ObjectError(Self, MT_DecompressFile, ERR_Write); goto exit; }

                     inputlen -= read.Result;
                     if (inputlen <= 0) break;
                     if (inputlen < SIZE_COMPRESSION_BUFFER) read.Length = inputlen;
                     else read.Length = SIZE_COMPRESSION_BUFFER;
                  }

                  if (error) goto exit;
               }
               else if ((filelist->DeflateMethod IS 8) AND (inflateInit2(&Self->prvZip, -MAX_WBITS) IS ERR_Okay)) {
                  inflateend = TRUE;

                  read.Buffer = Self->prvInput;
                  if (filelist->CompressedSize < SIZE_COMPRESSION_BUFFER) read.Length = filelist->CompressedSize;
                  else read.Length = SIZE_COMPRESSION_BUFFER;

                  err = ERR_Okay;
                  if ((error = Action(AC_Read, Self->FileIO, &read)) != ERR_Okay) goto exit;
                  if (read.Result <= 0) { error = ERR_Read; goto exit; }
                  inputlen = filelist->CompressedSize - read.Result;

                  Self->prvZip.next_in   = Self->prvInput;
                  Self->prvZip.avail_in  = read.Result;
                  Self->prvZip.next_out  = Self->prvOutput;
                  Self->prvZip.avail_out = SIZE_COMPRESSION_BUFFER;

                  /*** Keep loooping until Z_STREAM_END or an error is returned ***/

                  while (err IS ERR_Okay) {
                     err = inflate(&Self->prvZip, Z_SYNC_FLUSH);

                     if ((err != ERR_Okay) AND (err != Z_STREAM_END)) {
                        if (Self->prvZip.msg) DPrintF("!DecompressFile:","%s", Self->prvZip.msg);
                        error = ERR_Failed;
                        goto exit;
                     }

                     /*** Write out the decompressed data ***/

                     write.Buffer = Self->prvOutput;
                     write.Length = SIZE_COMPRESSION_BUFFER - Self->prvZip.avail_out;
                     if (Action(AC_Write, file, &write) != ERR_Okay) { error = ObjectError(Self, MT_DecompressFile, ERR_Write); goto exit; }

                     /*** Exit if all data has been written out ***/

                     if (Self->prvZip.total_out IS filelist->OriginalSize) break;

                     /*** Reset the output buffer ***/

                     Self->prvZip.next_out  = Self->prvOutput;
                     Self->prvZip.avail_out = SIZE_COMPRESSION_BUFFER;

                     /*** Read more data if necessary ***/

                     if ((Self->prvZip.avail_in <= 0) AND (inputlen > 0)) {
                        if (inputlen < SIZE_COMPRESSION_BUFFER) read.Length = inputlen;
                        else read.Length = SIZE_COMPRESSION_BUFFER;

                        if ((error = Action(AC_Read, Self->FileIO, &read)) != ERR_Okay) goto exit;
                        if (read.Result <= 0) { error = ERR_Read; break; }
                        inputlen -= read.Result;

                        Self->prvZip.next_in  = Self->prvInput;
                        Self->prvZip.avail_in = read.Result;
                     }
                  }

                  /*** Terminate the inflation process ***/

                  inflateEnd(&Self->prvZip);
                  inflateend = FALSE;
               }
            }

            /*** Give the file a date that matches the original ***/

            ActionTags(MT_SetFileDate, file, feedback.Year, feedback.Month, feedback.Day, feedback.Hour, feedback.Minute, feedback.Second);

            acFree(file);
            file = NULL;
         }

         if ((Self->Feedback) AND (feedback.Progress < 100)) {
            feedback.Progress = 100;
            Self->Feedback(Self, &feedback);
         }

         Self->prvFileIndex++;
      }
   }

   if (Self->OutputID) {
      StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "\nDecompression complete.");
      Print(Self, Self->prvOutput);
   }

exit:
   if (inflateend) inflateEnd(&Self->prvZip);
   if (file) acFree(file);

   if ((error IS ERR_Okay) AND (Self->prvFileIndex <= 0)) {
      DPrintF("DecompressFile:","No files matched the path \"%s\" from %d files.", Args->Path, Self->prvTotalFiles);
      error = ERR_Search;
   }

   StepBack();
   return(error);
}

/*****************************************************************************
** Action: Flush()
*/

static ERROR COMPRESS_Flush(struct Compression *Self, APTR Void)
{
   struct acWrite write;
   LONG length, zerror;
   WORD done;

   if (Self->Head.SubID) return(ERR_Okay);

   Self->prvZip.avail_in = 0;

   done = FALSE;

   for (;;) {
      /*** Write out any bytes that are still left in the compression buffer ***/

      if ((length = SIZE_COMPRESSION_BUFFER - Self->prvZip.avail_out) > 0) {
         write.Buffer = Self->prvOutput;
         write.Length = length;
         if (Action(AC_Write, Self->FileIO, &write) != ERR_Okay) return(ERR_Write);
         Self->prvZip.next_out  = Self->prvOutput;
         Self->prvZip.avail_out = SIZE_COMPRESSION_BUFFER;
      }

      if (done) break;

      zerror = deflate(&Self->prvZip, Z_FINISH);

      /*** Ignore the second of two consecutive flushes: ***/

      if ((!length) AND (zerror IS Z_BUF_ERROR)) zerror = ERR_Okay;

      done = ((Self->prvZip.avail_out != 0) OR (zerror IS Z_STREAM_END));

      if ((zerror != ERR_Okay) AND (zerror != Z_STREAM_END)) break;
   }

   acFlush(Self->FileIO);

   return(ERR_Okay);
}

/*****************************************************************************
** Compression: Free
*/

static ERROR COMPRESS_Free(struct Compression *Self, APTR Void)
{
   struct ZipFile *chain, *next;
   ULONG listsize, listoffset;
   UWORD filecount, namelen, extralen;

   /* Write the end of file signature for the zip file if we have been
   ** making modifications to it.
   */

   if ((Self->FileIO) AND (!Self->Head.SubID) AND (Self->prvCompressionCount > 0)) {
      if ((chain = Self->prvFiles)) {
         /*** Determine the start of the list offset ***/

         while (chain->Next) chain = (struct ZipFile *)chain->Next;
         ActionTags(AC_Seek, Self->FileIO, (DOUBLE)(chain->Offset + HEAD_NAMELEN), SEEK_START);
         namelen = ReadWord(Self->FileIO);
         extralen = ReadWord(Self->FileIO);
         ActionTags(AC_Seek, Self->FileIO, (DOUBLE)(chain->CompressedSize + namelen + extralen), SEEK_CURRENT);
         listoffset = chain->Offset + chain->CompressedSize + namelen + extralen + HEAD_LENGTH;

         /*** Write out the central directory file list ***/

         listsize   = 0;
         filecount  = 0;
         for (chain=Self->prvFiles; chain != NULL; chain=(struct ZipFile *)chain->Next) {
            wrb_word(chain->DeflateMethod, listdata+LIST_METHOD);
            wrb_long(chain->TimeStamp, listdata+LIST_TIMESTAMP);
            wrb_long(chain->CRC, listdata+LIST_CRC);
            wrb_long(chain->CompressedSize, listdata+LIST_COMPRESSEDSIZE);
            wrb_long(chain->OriginalSize, listdata+LIST_FILESIZE);
            wrb_word(chain->NameLen, listdata+LIST_NAMELEN);
            wrb_word(chain->CommentLen, listdata+LIST_COMMENTLEN);
            wrb_long(chain->Offset, listdata+LIST_OFFSET);
            wrb_long(chain->Flags, listdata+LIST_ATTRIB);
            Write(Self->FileIO, listdata, LIST_LENGTH);
            Write(Self->FileIO, chain->Name, chain->NameLen);
            if (chain->Comment) Write(Self->FileIO, chain->Comment, chain->CommentLen);

            listsize += LIST_LENGTH + chain->NameLen + chain->CommentLen;
            filecount++;
         }

         wrb_word(filecount,  taildata + TAIL_FILECOUNT1);
         wrb_word(filecount,  taildata + TAIL_FILECOUNT2);
         wrb_long(listsize,   taildata + TAIL_FILELISTSIZE);
         wrb_long(listoffset, taildata + TAIL_FILELISTOFFSET);
         Write(Self->FileIO, taildata, TAIL_LENGTH);
      }
      else SetFields(Self->FileIO, FID_Size|TLONG, 0, TAGEND);
   }

   /*** Free memory and resources ***/

   for (chain=Self->prvFiles; chain != NULL; chain=next) {
      next = (struct ZipFile *)chain->Next;
      FreeMemory(chain);
   }
   Self->prvFiles = NULL;

   if (Self->prvInput) {
      FreeMemory(Self->prvInput);
      Self->prvInput = NULL;
   }

   if (Self->prvOutput) {
      FreeMemory(Self->prvOutput);
      Self->prvOutput = NULL;
   }

   if (Self->FileIO) {
      acFree(Self->FileIO);
      Self->FileIO = NULL;
   }

   if (Self->Location) {
      FreeMemory(Self->Location);
      Self->Location = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Compression: HealthCheck
*/

static ERROR COMPRESS_HealthCheck(struct Compression *Self, APTR Void)
{
   /*** Test the validity of the compression data and repair it if possible ***/

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Init
Short: Prepares a compression object for usage.
###END###

*****************************************************************************/

static ERROR COMPRESS_Init(struct Compression *Self, APTR Void)
{
   struct ZipFile *entry, *lastentry;
   struct acRead read;
   STRING location;
   LONG type, result;

   GetField(Self, FID_Location, FT_STRING, &location);

   if (!location) {
      /* If no location has been set, assume that the developer
      ** only wants to use the buffer compression routines.
      */
      return(ERR_Okay);
   }
   else if (Self->Flags & CMF_NEW) {
      /*** If the NEW flag is set then create a new archive, destroying any file already at that location ***/

      if (CreateObject(ID_FILE, NF_CHILD, &Self->FileIO, NULL,
         FID_Location|TSTRING, location,
         FID_Flags|TLONG,      FL_READ|FL_WRITE|FL_NEW,
         TAGEND) IS ERR_Okay) {

         return(ERR_Okay);
      }
      else {
         if (Self->OutputID) {
            StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "Failed to create file \"%s\".", location);
            Print(Self, Self->prvOutput);
         }

         return(ObjectError(Self, ERH_Init, ERR_CreateObject));
      }
   }
   else if (CreateObject(ID_FILE, NF_CHILD, &Self->FileIO, NULL,
      FID_Location|TSTRING, location,
      FID_Flags|TLONG,      FL_READ|FL_APPROXIMATE| ((Self->Flags & CMF_READONLY) ? 0 : FL_WRITE),
      TAGEND) IS ERR_Okay) {

      /* Test the given location to see if it matches
      ** our supported file format (pkzip).
      */

      read.Buffer = Self->Header;
      read.Length = COMPRESS_LEN_HEADER;
      if (Action(AC_Read, Self->FileIO, &read) != ERR_Okay) return(ObjectError(Self, ERH_Init, ERR_Read));

      /*** If the file is empty then we will accept it as a zip file ***/

      if (read.Result IS 0) return(ERR_Okay);

      /*** Check for a pkzip header ***/

      if ((Self->Header[0] IS 0x50) AND (Self->Header[1] IS 0x4b) AND
          (Self->Header[2] IS 0x03) AND (Self->Header[3] IS 0x04)) {
         UBYTE strbuffer[256];

         if (ActionTags(AC_Seek, Self->FileIO, 0.0, SEEK_START) != ERR_Okay) return(ObjectError(Self, ERH_Init, ERR_Seek));

         lastentry = NULL;
         while ((type = ReadLong(Self->FileIO))) {
            if (type IS 0x04034b50) {
               /*** PKZIP file header entry detected ***/

               if (ActionTags(AC_Seek, Self->FileIO, (DOUBLE)HEAD_COMPRESSEDSIZE-4, SEEK_CURRENT) != ERR_Okay) return(ObjectError(Self, ERH_Init, ERR_Seek));

               struct {
                  ULONG compressedsize;
                  ULONG originalsize;
                  UWORD namelen;
                  UWORD extralen;
               } header;

               if (acRead(Self->FileIO, &header, sizeof(header), &result) != ERR_Okay) return(ObjectError(Self, ERH_Init, ERR_Read));

#ifndef REVERSE_BYTEORDER
               header.compressedsize = le32_cpu(header.compressedsize);
               header.originalsize   = le32_cpu(header.originalsize);
               header.namelen        = le16_cpu(header.namelen);
               header.extralen       = le16_cpu(header.extralen);
#endif

               if (ActionTags(AC_Seek, Self->FileIO, (DOUBLE)(header.compressedsize + header.namelen + header.extralen), SEEK_CURRENT) != ERR_Okay) return(ObjectError(Self, ERH_Init, ERR_Seek));
            }
            else if (type IS 0x02014b50) {
               /*** PKZIP list entry detected ***/

               struct {
                  UBYTE version;
                  UBYTE ostype;
                  UBYTE required_version;
                  UBYTE required_os;
                  UWORD flags;
                  UWORD deflatemethod;
                  ULONG timestamp;
                  ULONG crc32;
                  ULONG compressedsize;
                  ULONG originalsize;
                  UWORD namelen;
                  UWORD extralen;
                  UWORD commentlen;
                  UWORD diskno;
                  UWORD ifile;
                  ULONG attrib;
                  ULONG offset;
               } __attribute__((__packed__)) zipentry;

               if (acRead(Self->FileIO, &zipentry, sizeof(zipentry), &result) != ERR_Okay) return(ObjectError(Self, ERH_Init, ERR_Read));

#ifndef REVERSE_BYTEORDER
               zipentry.deflatemethod  = le16_cpu(zipentry.deflatemethod);
               zipentry.timestamp      = le32_cpu(zipentry.timestamp);
               zipentry.crc32          = le32_cpu(zipentry.crc32);
               zipentry.compressedsize = le32_cpu(zipentry.compressedsize);
               zipentry.originalsize   = le32_cpu(zipentry.originalsize);
               zipentry.namelen        = le16_cpu(zipentry.namelen);
               zipentry.extralen       = le16_cpu(zipentry.extralen);
               zipentry.commentlen     = le16_cpu(zipentry.commentlen);
               zipentry.diskno         = le16_cpu(zipentry.diskno);
               zipentry.ifile          = le16_cpu(zipentry.ifile);
               zipentry.attrib         = le32_cpu(zipentry.attrib);
               zipentry.offset         = le32_cpu(zipentry.offset);
#endif

               /* Note: Hidden memory is allocated to save time and resource tracking space
               ** when loading huge compressed files.
               */

               if (AllocMemory(sizeof(struct ZipFile) + zipentry.namelen + 1 + zipentry.commentlen + 1, MEM_DATA, &entry, NULL) != ERR_Okay) return(ObjectError(Self, ERH_Init, ERR_Memory));

               /*** Read the file name string ***/

               entry->Name = (STRING)(entry+1);
               if (acRead(Self->FileIO, entry->Name, zipentry.namelen, &result) IS ERR_Okay) {
                  entry->Name[zipentry.namelen] = 0;
               }
               else return(ObjectError(Self, ERH_Init, ERR_Read));

               /*** Read extra file information ***/

               if (zipentry.extralen > 0) {
                  read.Buffer = strbuffer;
                  read.Length = zipentry.extralen;
                  if (Action(AC_Read, Self->FileIO, &read) IS ERR_Okay) {

                  }
                  else return(ObjectError(Self, ERH_Init, ERR_Read));
               }

               /*** Read the file comment string, if any ***/

               if (zipentry.commentlen > 0) {
                  entry->Comment = entry->Name + zipentry.namelen + 1;
                  if (acRead(Self->FileIO, strbuffer, zipentry.commentlen, &result) IS ERR_Okay) {
                     CopyMemory(strbuffer, entry->Comment, zipentry.commentlen);
                     entry->Comment[zipentry.commentlen] = 0;
                  }
                  else return(ObjectError(Self, ERH_Init, ERR_Read));
               }

               /*** Build the file entry structure ***/

               entry->NameLen        = zipentry.namelen;
               entry->CommentLen     = zipentry.commentlen;
               entry->Comment        = NULL;
               entry->CompressedSize = zipentry.compressedsize;
               entry->OriginalSize   = zipentry.originalsize;
               entry->DeflateMethod  = zipentry.deflatemethod;
               entry->TimeStamp      = zipentry.timestamp;
               entry->CRC            = zipentry.crc32;
               entry->Offset         = zipentry.offset;
               if (zipentry.ostype IS ZIP_ATHENE) entry->Flags = zipentry.attrib;

               /*** Chain management ***/

               if (lastentry) {
                  entry->Prev = (struct CompressedFile *)lastentry;
                  lastentry->Next = (struct CompressedFile *)entry;
               }
               else Self->prvFiles = entry;

               lastentry = entry;
            }
            else if (type IS 0x06054b50) {
               /*** PKZIP end of file directory signature detected ***/

               break; /* End the loop */
            }
            else {
               /*** Unrecognised PKZIP data ***/
               DPrintF("@Init:","Unrecognised PKZIP entry $%.8x in the central directory.", type);
               return(ERR_BadData);
            }
         }

         DPrintF("Compression:","The file has been successfully scanned.");

         return(ERR_Okay);
      }
      else return(ERR_NoSupport);
   }
   else if (Self->Flags & CMF_CREATEFILE) {
      /*** Create a new file if the requested location does not exist ***/

      DPrintF("Init:","[Compression] Creating a new file because the location does not exist.");

      if (CreateObject(ID_FILE, NF_CHILD, &Self->FileIO, NULL,
         FID_Location|TSTRING, location,
         FID_Flags|TLONG,      FL_READ|FL_WRITE|FL_NEW,
         TAGEND) IS ERR_Okay) {

         return(ERR_Okay);
      }
      else {
         if (Self->OutputID) {
            StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "Failed to create file \"%s\".", location);
            Print(Self, Self->prvOutput);
         }

         return(ObjectError(Self, ERH_Init, ERR_CreateObject));
      }
   }
   else {
      StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "Failed to open \"%s\".", location);
      Print(Self, Self->prvOutput);
      return(ObjectError(Self, ERH_Init, ERR_File));
   }
}

/*****************************************************************************
** Compression: NewObject
*/

static ERROR COMPRESS_NewObject(struct Compression *Self, APTR Void)
{
   if (AllocMemory(SIZE_COMPRESSION_BUFFER, MEM_DATA, &Self->prvOutput, NULL) IS ERR_Okay) {
      if (AllocMemory(SIZE_COMPRESSION_BUFFER, MEM_DATA, &Self->prvInput, NULL) IS ERR_Okay) {
         Self->CompressionLevel = 60; /* 60% compression by default */
         Self->Permissions = PERMIT_READ|PERMIT_WRITE|PERMIT_GROUP_READ|PERMIT_GROUP_WRITE;
         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_NewObject, ERR_AllocMemory));
   }
   else return(ObjectError(Self, ERH_NewObject, ERR_AllocMemory));
}

/*****************************************************************************

###METHOD###
Name:  RemoveFile
Short: Deletes one or more files from a compression object.

###DESCRIPTION###
<p>This method deletes compressed files from a compression object.  If the
file is in a directory then you must specify the complete path in conjunction
with the file name.  Wild cards are accepted if you want to delete multiple
files.  A Path setting of &quot;*&quot; will delete an archive's entire
contents, while a more conservative Path of &quot;documents/ *&quot; would
delete all files and directories under the documents path.  Directories
can be declared using either the back-slash or the forward-slash
characters.</p>

<p>Depending on internal optimisation techniques, the compressed file may not
shrink from deletions until the compression object is closed or the Flush
action is called.</p>

###INPUT###
STRING/Path: The full path name of the file to delete from the archive.

###RESULT###
ERR_Okay: The file was successfully deleted.
ERR_Args: Invalid arguments were detected.

###END###

*****************************************************************************/

static ERROR COMPRESS_RemoveFile(struct Compression *Self, struct mtRemoveFile *Args)
{
   struct ZipFile *filelist;
   ERROR error;

   if ((!Args) OR (!Args->Path)) return(ObjectError(Self, MT_RemoveFile, ERR_Args));

   if (Self->Head.SubID) return(ERR_NoSupport);

   /* Search for the file(s) in our archive that match the given name
   ** and delete them.
   */

   DPrintF("RemoveFile()","[Compression:%d] %s", Self->Head.UniqueID, Args->Path);

   filelist = Self->prvFiles;
   while (filelist) {
      if (StrCompare(Args->Path, filelist->Name, NULL, STR_WILDCARD) IS ERR_Okay) {
         /*** Delete the file from the archive ***/

         if (Self->OutputID) {
            StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "Removing file \"%s\".", filelist->Name);
            Print(Self, Self->prvOutput);
         }

         if ((error = RemoveFile(Self, &filelist)) != ERR_Okay) return(error);
      }
      else filelist = (struct ZipFile *)filelist->Next;
   }

   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   CompressionLevel
Short:  The compression level to use when compressing data.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The level of compression that is used when compressing data is determined
by the value in this field.  Values range between 0 for no compression and
100 for maximum compression.  The speed of compression decreases with higher
values but this will give you improved size reduction.</p>

###END###

****************************************************************************/

static ERROR SET_CompressionLevel(struct Compression *Self, LONG *Value)
{
   Self->CompressionLevel = *Value;
   if (Self->CompressionLevel < 0) Self->CompressionLevel = 0;
   else if (Self->CompressionLevel > 100) Self->CompressionLevel = 100;
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Flags
Short:  Special flags.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>You can specify flags in this field that will affect the way that a
compression object is managed.  Valid flags are:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">PASSWORD</td><td>This flag is automatically set when it is possible to detect compression files that have been encrypted.  You can test it to see if a compression object requires a password to decompress the data.</td></tr>
<tr><td width="10%">NEW</td><td>If a compression file is being created from scratch (i.e. you will not be loading any compression data), set the NEW flag so that the initialisation process will create a compression file from scratch.</td></tr>
<tr><td width="10%">READONLY</td><td>If you are opening an existing compression file and do not intend to add new files to it, we recommend setting the READONLY flag.  If this flag is not set, opening the compression file is more likely to fail if the user does not have write access to it.</td></tr>
</table>

###END###

****************************************************************************/

/****************************************************************************

###FIELD###
Name:   FileList
Short:  A chained list of all files in a compression object.
Type:   struct CompressedFile *
Status: Read

###DESCRIPTION###
<p>The FileList field provides a list of all the files that are contained
in a compresion object, including details such as the date and original
size of each file.  The FileList structure is arranged as follows:</p>

<pre>
   struct CompressedFile {
      struct CompressedFile *Next;
      struct CompressedFile *Prev;
      STRING Name;
      ULONG  CompressedSize;
      ULONG  OriginalSize;
      LONG   Year;
      UBYTE  Month;
      UBYTE  Day;
      UBYTE  Hour;
      UBYTE  Minute;
   };
</pre>

<p>The field names are self explanatory.  To peruse the file list, use the
Next and Prev fields to move between each record.  The list ends when the
Next field is set to NULL and the first record is indicated with a Prev
field of NULL.</p>

###END###

****************************************************************************/

static ERROR GET_FileList(struct Compression *Self, struct CompressedFile **List)
{
   *List = (struct CompressedFile *)Self->prvFiles;
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Header
Short:  The first 32 bytes of a compression object's file header.
Type:   UBYTE *
Status: Read

###DESCRIPTION###
<p>This field is only of use to sub-classes that need to examine the first
32 bytes of a compressed file's header.</p>

###END###

****************************************************************************/

static ERROR GET_Header(struct Compression *Self, UBYTE **Header)
{
   *Header = Self->Header;
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:     Location
Synonyms: Src
Short:    Set this field if the compression data originates from, or is to be saved to a file source.
Type:     STRING
Status:   Read/Write

###DESCRIPTION###
<p>When using a compression object to load or create a new compressed file,
you must set the Location field.  If for example you wish to examine the
contents of a zip file in the root directory, you might set this field to
&quot;athene:data.zip&quot;.</p>

###END###

****************************************************************************/

static ERROR GET_Location(struct Compression *Self, STRING *Value)
{
   if (Self->Location) { *Value = Self->Location; return(ERR_Okay); }
   else return(ERR_NoData);
}

static ERROR SET_Location(struct Compression *Self, STRING Value)
{
   LONG i;

   if (Self->Location) { FreeMemory(Self->Location); Self->Location = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->Location, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Location[i] = Value[i];
         Self->Location[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Output
Short:  Resulting messages will be sent to the object referred to in this field.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>If this field is set to a valid ObjectID, text messages will be sent to
that object when the compression object is used.  This can be helpful for
notifying the user of the results of compression, decompression and removal
of files.</p>

<p>The object receiving the message must be capable of understanding text
sent via data channels.  In most cases it is recommended that a Text
object is used for this purpose.</p>

###END###

****************************************************************************/

/****************************************************************************

###FIELD###
Name:   Password
Short:  Required if the compressed data uses a password.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>If you would like to set a Password for a new compression object, or if
an existing compression object requires a password, you will need to set the
Password field.  The string that you supply must be null-terminated and not
more than 128 bytes in length.</p>

<p>It is recommended that you set the Password before or immediately after
initialisation.  If want to change the password or add a password to an
existing compression object, you will need to create a new compresion object
and transfer the existing data across to it.</p>

###END###

****************************************************************************/

static ERROR GET_Password(struct Compression *Self, STRING *Value)
{
   *Value = Self->Password;
   return(ERR_Okay);
}

static ERROR SET_Password(struct Compression *Self, STRING Value)
{
   LONG i;

   if ((Value) AND (*Value)) {
      for (i=0; (i < LEN_PASSWORD-1) AND (Value[i]); i++) Self->Password[i] = Value[i];
      Self->Password[i] = 0;
      Self->Flags |= CMF_PASSWORD;
   }
   else Self->Password[0] = 0;

   return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   Size
Short:  The byte size of a compressed file.
Type:   LARGE
Status: Get

###DESCRIPTION###
<p>You can learn the compressed size of a compression object's associated
data file by reading the Size field.  The total includes the size of each
compressed file as well as the size of headers and other associated data
that may be included in the archive.</p>

###END###

****************************************************************************/

static ERROR GET_Size(struct Compression *Self, LARGE *Value)
{
   *Value = 0;

   if (Self->FileIO) {
      return(GetField(Self->FileIO, FID_Size, FT_LARGE, Value));
   }
   else return(ERR_Okay);
}

/****************************************************************************

###FIELD###
Name:   UncompressedSize
Short:  Indicates the total uncompressed size of all files in a compression object.
Type:   LARGE
Status: Get

###DESCRIPTION###
<p>If you would like to know the total number of bytes that have been
compressed into a compression object, read this field.  This will tell you
the maximium byte count used if every file were to be decompressed.  Header
and tail information that may identify the compressed data is not included
in the total.</p>

###END###

****************************************************************************/

static ERROR GET_UncompressedSize(struct Compression *Self, LARGE *Value)
{
   struct ZipFile *filelist;
   LARGE size;

   size = NULL;
   for (filelist=Self->prvFiles; filelist; filelist=(struct ZipFile *)filelist->Next) {
      size += filelist->OriginalSize;
   }
   *Value = size;
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: Print()
*/

static void Print(struct Compression *Self, STRING Buffer)
{
   struct acDataChannel channel;

   if (Self->OutputID) {
      channel.ObjectID = Self->Head.UniqueID;
      channel.DataType = DATA_TEXT;
      channel.Version  = 1;
      channel.Buffer   = Buffer;
      for (channel.Size=0; Buffer[channel.Size]; channel.Size++);
      channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &channel);
   }
   else DPrintF("Compression:", Buffer);
}

/*****************************************************************************
** Internal: CompressDirectory()
*/

static ERROR CompressDirectory(struct Compression *Self, STRING Location, STRING Path)
{
   struct ZipFile *entry, *chain, *fileexists;
   struct FileInfo info;
   struct CompressionFeedback feedback;
   struct FileInfo *list;
   LONG i, j, len, pathlen, result;
   ULONG dataoffset;
   ERROR error;
   UWORD namelen, extralen;

   DPrintF("4CompressFolder()","[Compression:%d] Compressing folder \"%s\" to \"%s\"", Self->Head.UniqueID, Location, Path);

   if ((!Self) OR (!Location)) return(ERR_Args);
   if (!Path) Path = "";

   if (GetFileInfo(Location, &info, sizeof(info)) != ERR_Okay) {
      DPrintF("@CompressFolder:","Failed to get folder info.");
      return(ERR_Failed);
   }

   if ((info.Flags & RDF_LINK) AND (!(Self->Flags & CMF_NOLINKS))) {
      DPrintF("CompressFolder:","Folder is a link.");
      error = CompressFile(Self, Location, Path, TRUE);
      return(error);
   }

   if (Self->OutputID) {
      StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "  Compressing folder \"%s\".", Location);
      Print(Self, Self->prvOutput);
   }

   /*** Send feedback if requested to do so ***/

   if (Self->Feedback) {
      feedback.FeedbackID     = FDB_CompressFile;
      feedback.Index          = Self->prvFileIndex;
      feedback.Path           = Location;
      feedback.Destination    = Path;
      feedback.OriginalSize   = 0;
      feedback.CompressedSize = 0;
      feedback.Progress       = 100;

      result = Self->Feedback(Self, &feedback);

      if (result IS CFR_Abort) {
         Self->prvFileIndex++;
         return(ERR_Cancelled);
      }
      else if (result IS CFR_Skip) {
         Self->prvFileIndex++;
         return(ERR_Okay);
      }
   }
   else Self->prvFileIndex++;

   /*** Clear default variables ***/

   fileexists = NULL;
   entry      = NULL;
   error      = ERR_Failed;

   for (pathlen=0; Path[pathlen]; pathlen++);

   if (pathlen > 0) {
      /*** Seek to the position at which this new directory entry will be added ***/

      dataoffset = 0;
      if ((chain = Self->prvFiles)) {
         while (chain->Next) chain = (struct ZipFile *)chain->Next;
         if (ActionTags(AC_Seek, Self->FileIO, (DOUBLE)(chain->Offset + HEAD_NAMELEN), SEEK_START) != ERR_Okay) return(ObjectError(Self, MT_CompressFile, ERR_Seek));
         namelen    = ReadWord(Self->FileIO);
         extralen   = ReadWord(Self->FileIO);
         dataoffset = chain->Offset + HEAD_LENGTH + namelen + extralen + chain->CompressedSize;
      }

      if (ActionTags(AC_Seek, Self->FileIO, (DOUBLE)dataoffset, SEEK_START) != ERR_Okay) goto exit;

      /*** If a matching file name already exists in the archive, make a note of its position ***/

      for (fileexists=Self->prvFiles; fileexists; fileexists=(struct ZipFile *)fileexists->Next) {
         if (StrCompare(fileexists->Name, Location, 0, STR_MATCHLENGTH) IS ERR_Okay) break;
      }

      /* Allocate the file entry structure and set up some initial
      ** variables.
      */

      if (AllocMemory(sizeof(struct ZipFile) + pathlen + 1, MEM_DATA, &entry, NULL) != ERR_Okay) {
         return(ERR_AllocMemory);
      }

      entry->Name = (STRING)(entry+1);
      entry->NameLen = pathlen;
      for (i=0; i < pathlen; i++) entry->Name[i] = Path[i];
      entry->Name[i] = 0;
      entry->CRC        = 0;
      entry->Offset     = dataoffset;
      entry->CompressedSize = 0;
      entry->OriginalSize   = 0;
      entry->DeflateMethod  = 0;

      /*** Convert the file date stamp into a DOS time stamp for zip ***/

      if (info.Time.Year < 1980) entry->TimeStamp = 0x00210000;
      else entry->TimeStamp = ((info.Time.Year-1980)<<25) | (info.Time.Month<<21) | (info.Time.Day<<16) | (info.Time.Hour<<11) | (info.Time.Minute<<5) | (info.Time.Second>>1);

      /*** Write the compression file entry ***/

      if (ActionTags(AC_Seek, Self->FileIO, (DOUBLE)entry->Offset, SEEK_START) != ERR_Okay) goto exit;
      wrb_word(entry->DeflateMethod, header + HEAD_DEFLATEMETHOD);
      wrb_long(entry->TimeStamp, header + HEAD_TIMESTAMP);
      wrb_long(entry->CRC, header + HEAD_CRC);
      wrb_long(entry->CompressedSize, header + HEAD_COMPRESSEDSIZE);
      wrb_long(entry->OriginalSize, header + HEAD_FILESIZE);
      wrb_word(entry->NameLen, header + HEAD_NAMELEN);
      if (Write(Self->FileIO, header, HEAD_LENGTH) != HEAD_LENGTH) goto exit;
      if (Write(Self->FileIO, entry->Name, entry->NameLen) != entry->NameLen) goto exit;

      /*** Add the entry to the file chain ***/

      if ((chain = Self->prvFiles)) {
         while (chain->Next) chain = (struct ZipFile *)chain->Next;
         entry->Prev = (struct CompressedFile *)chain;
         chain->Next = (struct CompressedFile *)entry;
      }
      else Self->prvFiles = entry;

      /*** If this new data replaces an existing directory, remove the old directory now ***/

      if (fileexists) RemoveFile(Self, &fileexists);

      Self->prvCompressionCount++;
   }

   /*** Enter the directory and compress its contents ***/

   if (ReadDirectory(Location, RDF_FILE|RDF_DIRECTORY|RDF_QUALIFY, &list) IS ERR_Okay) {
      struct FileInfo *scan;

      /*** Get the length of the location and path arguments ***/

      for (len=0; Location[len]; len++);
      for (pathlen=0; Path[pathlen]; pathlen++);

      /*** Recurse for each directory in the list ***/

      for (scan=list; scan; scan=scan->Next) {
         if ((scan->Flags & RDF_DIRECTORY) AND (!(scan->Flags & RDF_LINK))) {
            for (j=0; scan->Name[j]; j++);

            UBYTE location[len+j+1];
            UBYTE path[pathlen+j+1];

            StrFormat(location, sizeof(location), "%s%s", Location, scan->Name);
            StrFormat(path, sizeof(path), "%s%s", Path, scan->Name);
            CompressDirectory(Self, location, path);
         }
      }

      for (scan=list; scan; scan=scan->Next) {
         if (scan->Flags & (RDF_FILE|RDF_LINK)) {
            for (j=0; scan->Name[j]; j++);

            UBYTE location[len+j+1];

            j = StrCopy(Location, location, sizeof(location));
            StrCopy(scan->Name, location + j, sizeof(location) - j);
            CompressFile(Self, location, Path, (scan->Flags & RDF_LINK) ? TRUE : FALSE);
         }
      }

      FreeMemory(list);
   }

   return(ERR_Okay);

exit:
   if (entry) {
      FreeFromChain((struct CompressedFile *)entry, (struct CompressedFile *)Self->prvFiles);
      FreeMemory(entry);
   }

   return(error);
}

/*****************************************************************************
** Internal: CompressFile()
*/

static ERROR CompressFile(struct Compression *Self, STRING Location, STRING Path, BYTE Link)
{
   struct acRead read;
   struct acWrite write;
   struct CompressionFeedback feedback;
   struct ZipFile *entry, *chain, *fileexists;
   struct FileTime *time;
   struct File *file;
   STRING symlink;
   UBYTE filename[512];
   WORD level;
   UWORD namelen, extralen;
   LONG i, len, result, permissions;
   UBYTE deflateend;
   ULONG dataoffset;
   ERROR error;

   if ((!Self) OR (!Location) OR (!Path)) return(ERR_Args);

   DPrintF("6CompressFile()","[Compression:%d] Compressing file \"%s\" to \"%s\"", Self->Head.UniqueID, Location, Path);

   /*** Clear default variables ***/

   file       = NULL;
   filename[0] = 0;
   fileexists = NULL;
   entry      = NULL;
   symlink    = NULL;
   error      = ERR_Failed;
   deflateend = FALSE;

   /*** Open the source file for reading only ***/

   if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
      FID_Location|TSTRING, Location,
      FID_Flags|TLONG,      (Link IS TRUE) ? NULL : FL_READ,
      TAGEND) != ERR_Okay) {
      if (Self->OutputID) {
         StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "  Error opening file \"%s\".", Location);
         Print(Self, Self->prvOutput);
      }
      error = ObjectError(Self, MT_CompressFile, ERR_OpenFile);
      goto exit;
   }

   if ((Link) AND (!(file->Flags & FL_SYMLINK))) {
      DPrintF("@CompressFile:","Internal Error: Expected a link, but the file is not.");
      error = ERR_Failed;
      goto exit;
   }

   /*** Determine the name that will be used for storing this file ***/

   for (i=0; Location[i]; i++);
   while ((i > 0) AND (Location[i-1] != ':') AND (Location[i-1] != '/') AND (Location[i-1] != '\\')) i--;
   if (Path) {
      len = StrCopy(Path, filename, sizeof(filename));
      StrCopy(Location+i, filename+len, sizeof(filename)-len);
   }
   else StrCopy(Location+i, filename, sizeof(filename));

   if (Link) {
      for (i=0; filename[i]; i++);
      if (filename[i-1] IS '/') filename[i-1] = 0;
   }

   /*** Send feedback if requested to do so ***/

   if (Self->Feedback) {
      feedback.FeedbackID     = FDB_CompressFile;
      feedback.Index          = Self->prvFileIndex;
      feedback.Path           = Location;
      feedback.Destination    = filename;
      GetField(file, FID_Size, FT_LONG, &feedback.OriginalSize);
      feedback.CompressedSize = 0;
      feedback.Progress       = 0;
      result = Self->Feedback(Self, &feedback);

      if (result IS CFR_Abort) { error = ERR_Cancelled; goto exit; }
      else if (result IS CFR_Skip) { error = ERR_Okay; goto exit; }
   }

   /*** Send informative output to the user ***/

   if (Self->OutputID) {
      StrFormat(Self->prvOutput, SIZE_COMPRESSION_BUFFER, "  Compressing file \"%s\".", Location);
      Print(Self, Self->prvOutput);
   }

   /*** Seek to the position at which this new file will be added ***/

   dataoffset = 0;
   if ((chain = Self->prvFiles)) {
      while (chain->Next) chain = (struct ZipFile *)chain->Next;
      if (ActionTags(AC_Seek, Self->FileIO, (DOUBLE)(chain->Offset + HEAD_NAMELEN), SEEK_START) != ERR_Okay) return(ObjectError(Self, MT_CompressFile, ERR_Seek));
      namelen    = ReadWord(Self->FileIO);
      extralen   = ReadWord(Self->FileIO);
      dataoffset = chain->Offset + HEAD_LENGTH + namelen + extralen + chain->CompressedSize;
   }

   if (ActionTags(AC_Seek, Self->FileIO, (DOUBLE)dataoffset, SEEK_START) != ERR_Okay) goto exit;

   /*** Initialise the compression algorithm ***/

   Self->prvCompressionCount++;

   level = Self->CompressionLevel / 10;
   if (level < 0) level = 0;
   else if (level > 9) level = 9;

   Self->prvZip.next_in   = 0;
   Self->prvZip.avail_in  = 0;
   Self->prvZip.next_out  = 0;
   Self->prvZip.avail_out = 0;
   Self->prvZip.total_in  = 0;
   Self->prvZip.total_out = 0;

   if (deflateInit2(&Self->prvZip, level, Z_DEFLATED, -MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY) IS ERR_Okay) {
      deflateend = TRUE;
      Self->prvZip.next_out  = Self->prvOutput;
      Self->prvZip.avail_out = SIZE_COMPRESSION_BUFFER;
   }
   else {
      error = ERR_Failed;
      goto exit;
   }

   /*** If a matching file name already exists in the archive, make a note of its position ***/

   for (fileexists=Self->prvFiles; fileexists; fileexists=(struct ZipFile *)fileexists->Next) {
      if (StrCompare(fileexists->Name, filename, 0, STR_MATCHLENGTH) IS ERR_Okay) break;
   }

   /* Allocate the file entry structure and set up some initial
   ** variables.
   */

   for (len=0; filename[len]; len++);
   if (AllocMemory(sizeof(struct ZipFile) + len + 1, MEM_DATA, &entry, NULL) IS ERR_Okay) {
      entry->Name = (STRING)(entry+1);
      entry->NameLen = len;
      for (i=0; i < entry->NameLen; i++) entry->Name[i] = filename[i];
      entry->Name[i] = 0;
      entry->CRC        = 0;
      entry->Offset     = dataoffset;
      entry->Comment    = 0;
      entry->CommentLen = 0;

      if ((!(Self->Flags & CMF_NOLINKS)) AND (file->Flags & FL_SYMLINK)) {
         if (GetField(file, FID_Link, FT_STRING, &symlink) IS ERR_Okay) {
            DPrintF("CompressFile:","Note: File \"%s\" is a symbolic link to \"%s\"", filename, symlink);
            entry->Flags |= ZIP_LINK;
         }
      }

      /*** Convert the file date stamp into a DOS time stamp for zip ***/

      if (GetField(file, FID_Date, FD_POINTER, &time) IS ERR_Okay) {
         if (time->Year < 1980) entry->TimeStamp = 0x00210000;
         else entry->TimeStamp = ((time->Year-1980)<<25) | (time->Month<<21) | (time->Day<<16) | (time->Hour<<11) | (time->Minute<<5) | (time->Second>>1);
      }
      else entry->TimeStamp = NULL;
   }
   else {
      error = ERR_AllocMemory;
      goto exit;
   }

   if (GetField(file, FID_Permissions, FT_LONG, &permissions) IS ERR_Okay) {
      if (permissions & PERMIT_USER_READ) entry->Flags |= ZIP_UREAD;
      if (permissions & PERMIT_GROUP_READ) entry->Flags |= ZIP_GREAD;
      if (permissions & PERMIT_OTHERS_READ) entry->Flags |= ZIP_OREAD;

      if (permissions & PERMIT_USER_WRITE) entry->Flags |= ZIP_UWRITE;
      if (permissions & PERMIT_GROUP_WRITE) entry->Flags |= ZIP_GWRITE;
      if (permissions & PERMIT_OTHERS_WRITE) entry->Flags |= ZIP_OWRITE;

      if (permissions & PERMIT_USER_EXEC) entry->Flags |= ZIP_UEXEC;
      if (permissions & PERMIT_GROUP_EXEC) entry->Flags |= ZIP_GEXEC;
      if (permissions & PERMIT_OTHERS_EXEC) entry->Flags |= ZIP_OEXEC;
   }

   entry->Flags &= 0xffffff00; /* Do not write anything to the low order bits, they have meaning exclusive to MSDOS */

   /* Skip over the PKZIP header that will be written for this file (we will
   ** be updating the header later).
   */

   if (Write(Self->FileIO, NULL, HEAD_LENGTH + entry->NameLen + entry->CommentLen) != HEAD_LENGTH + entry->NameLen + entry->CommentLen) goto exit;

   /* Specify the limitations of our buffer so that the compression
   ** routine doesn't overwrite its boundaries.  Then start the
   ** compression of the input file.
   */

   if (entry->Flags & ZIP_LINK) {
      /*** Compress the symbolic link to the zip file, rather than the data ***/
      len = StrLength(symlink);
      Self->prvZip.next_in  = symlink;
      Self->prvZip.avail_in = len;
      Self->prvZip.next_out = Self->prvOutput;
      Self->prvZip.avail_out = SIZE_COMPRESSION_BUFFER;
      if (deflate(&Self->prvZip, Z_NO_FLUSH) != ERR_Okay) {
         DPrintF("!CompressFile:","[Compression] Failure during data compression.");
         goto exit;
      }
      entry->CRC = crc32(entry->CRC, symlink, len);
   }
   else {
      read.Buffer = Self->prvInput;
      read.Length = SIZE_COMPRESSION_BUFFER;
      while ((Action(AC_Read, file, &read) IS ERR_Okay) AND (read.Result > 0)) {
         Self->prvZip.next_in   = Self->prvInput;
         Self->prvZip.avail_in  = read.Result;

         while (Self->prvZip.avail_in) {
            if (!Self->prvZip.avail_out) {
               /*** Write out the compression buffer because it is at capacity ***/
               write.Buffer = Self->prvOutput;
               write.Length = SIZE_COMPRESSION_BUFFER;
               Action(AC_Write, Self->FileIO, &write);

               /*** Reset the compression buffer ***/
               Self->prvZip.next_out = Self->prvOutput;
               Self->prvZip.avail_out = SIZE_COMPRESSION_BUFFER;

               if (Self->Feedback) {
                  feedback.CompressedSize = Self->prvZip.total_out;
                  feedback.Progress       = feedback.CompressedSize * 100 / feedback.OriginalSize;
                  Self->Feedback(Self, &feedback);
               }
            }

            if (deflate(&Self->prvZip, Z_NO_FLUSH) != ERR_Okay) {
               DPrintF("!CompressFile:","[Compression] Failure during data compression.");
               goto exit;
            }
         }

         entry->CRC = crc32(entry->CRC, Self->prvInput, read.Result);
      }
   }

   if (acFlush(Self) != ERR_Okay) goto exit;
   deflateEnd(&Self->prvZip);
   deflateend = FALSE;

   /*** Finalise entry details ***/

   entry->CompressedSize = Self->prvZip.total_out;
   entry->OriginalSize   = Self->prvZip.total_in;

   if (entry->OriginalSize > 0) {
      entry->DeflateMethod = 8;
   }
   else {
      entry->DeflateMethod = 0;
      entry->CompressedSize = 0;
   }

   if ((chain = Self->prvFiles)) {
      while (chain->Next) chain = (struct ZipFile *)chain->Next;
      entry->Prev = (struct CompressedFile *)chain;
      chain->Next = (struct CompressedFile *)entry;
   }
   else Self->prvFiles = entry;

   /* Update the header that we earlier wrote for our file entry.  Note
   ** that the header stores only some of the file information, most of it
   ** is stored in the compressed file list at the end of the zip file.
   */

   if (ActionTags(AC_Seek, Self->FileIO, (DOUBLE)entry->Offset, SEEK_START) != ERR_Okay) goto exit;
   wrb_word(entry->DeflateMethod, header + HEAD_DEFLATEMETHOD);
   wrb_long(entry->TimeStamp, header + HEAD_TIMESTAMP);
   wrb_long(entry->CRC, header + HEAD_CRC);
   wrb_long(entry->CompressedSize, header + HEAD_COMPRESSEDSIZE);
   wrb_long(entry->OriginalSize, header + HEAD_FILESIZE);
   wrb_word(entry->NameLen, header + HEAD_NAMELEN);
   if (Write(Self->FileIO, header, HEAD_LENGTH) != HEAD_LENGTH) goto exit;
   if (Write(Self->FileIO, entry->Name, entry->NameLen) != entry->NameLen) goto exit;

   /*** Send updated feedback if necessary ***/

   if ((Self->Feedback) AND (feedback.Progress < 100)) {
      feedback.CompressedSize = entry->CompressedSize;
      feedback.Progress       = 100;
      Self->Feedback(Self, &feedback);
   }

   /*** If this new data replaces an existing file, remove the old file now ***/

   if (fileexists) RemoveFile(Self, &fileexists);

   acFree(file);
   Self->prvFileIndex++;
   return(ERR_Okay);

exit:
   if (deflateend IS TRUE) deflateEnd(&Self->prvZip);

   if (entry) {
      FreeFromChain((struct CompressedFile *)entry, (struct CompressedFile *)Self->prvFiles);
      FreeMemory(entry);
   }

   if (file) acFree(file);

   Self->prvFileIndex++;
   return(error);
}

/*****************************************************************************
** Internal: RemoveFile()
*/

ERROR RemoveFile(struct Compression *Self, struct ZipFile **File)
{
   struct acRead read;
   struct acWrite write;
   struct ZipFile *next, *file;
   DOUBLE currentpos, writepos;
   LONG chunksize;
   UWORD namelen, extralen;

   file = *File;

   DPrintF("RemoveFile:","[Compression:%d] Deleting \"%s\"", Self->Head.UniqueID, file->Name);

   /* Seek to the end of the compressed file.  We are going to delete
   ** the file by shifting all the data after the file to the start of
   ** the file's position.
   */

   if (ActionTags(AC_Seek, Self->FileIO, (DOUBLE)(file->Offset + HEAD_NAMELEN), SEEK_START) != ERR_Okay) return(ObjectError(Self, MT_RemoveFile, ERR_Seek));
   namelen    = ReadWord(Self->FileIO);
   extralen   = ReadWord(Self->FileIO);
   chunksize  = HEAD_LENGTH + namelen + extralen + file->CompressedSize;
   currentpos = file->Offset + chunksize;
   if (ActionTags(AC_Seek, Self->FileIO, currentpos, SEEK_START) != ERR_Okay) return(ObjectError(Self, MT_RemoveFile, ERR_Seek));

   writepos = file->Offset;

   read.Buffer = Self->prvInput;
   read.Length = SIZE_COMPRESSION_BUFFER;
   while ((Action(AC_Read, Self->FileIO, &read) IS ERR_Okay) AND (read.Result > 0)) {
      if (ActionTags(AC_Seek, Self->FileIO, writepos, SEEK_START) != ERR_Okay) return(ObjectError(Self, MT_RemoveFile, ERR_Seek));
      write.Buffer = Self->prvInput;
      write.Length = read.Result;
      if (Action(AC_Write, Self->FileIO, &write) != ERR_Okay) return(ObjectError(Self, MT_RemoveFile, ERR_Write));
      writepos += write.Result;

      currentpos += read.Result;
      if (ActionTags(AC_Seek, Self->FileIO, currentpos, SEEK_START) != ERR_Okay) return(ObjectError(Self, MT_RemoveFile, ERR_Seek));
   }

   SetField(Self->FileIO, FID_Size, FT_DOUBLE, &writepos);

   /*** Remove the file reference from the chain ***/

   next = (struct ZipFile *)file->Next;
   FreeFromChain((struct CompressedFile *)file, (struct CompressedFile *)Self->prvFiles);
   FreeMemory(file);

   /*** Adjust the offset of files that were ahead of this one ***/

   for (file=next; file != NULL; file=(struct ZipFile *)file->Next) file->Offset -= chunksize;

   *File = next;
   return(ERR_Okay);
}


