/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Ellipse
Version:   1.0
ID:        ID_ELLIPSE
Category:  GUI
Date:      March 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2003.  All rights reserved.
Keywords:  ellipse, circle, graphics, arc, elliptical, circular
Short:     Draws elliptical and circular shapes.

###DESCRIPTION###
<p>The Ellipse class is used to draw elliptical shapes and circles to
graphics areas.  All you need to specify are the dimensions of the ellipse
and its colour.  Here is a DML based example:</p>

<pre>
  &lt;ellipse x="10%" y="10%" width="80" height="60" colour="#6090a0"/&gt;
</pre>

<p>If you want to draw an unfilled ellipse, set the Border field as opposed
to the Colour.  Beyond this, there is little else that you need to know about
the Ellipse class - but check the documented fields if you need more
detial.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/functions/strtoint.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_ELLIPSE 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_ELLIPSE, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2003. All rights reserved.",
   "March 2003",
   "Ellipse"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR EllipseClass = 0;

/****************************************************************************
** Ellipse class.
*/

struct Ellipse {
   OBJECT_HEADER
   OBJECTID DrawableID;
   LONG     Frame;
   LONG     Thickness;

   PRIVATE_FIELDS
   struct RGB Highlight;
   struct RGB Shadow;
   struct RGB ColourRGB;
   struct BitmapPen *Pen;
   LONG  Dimensions;
   LONG  Flags;
   FLOAT XCoord;
   FLOAT YCoord;
   FLOAT Width;
   FLOAT Height;
   FLOAT XOffset;
   FLOAT YOffset;
   WORD  TotalFrames;
   WORD  *Frames;
};

/*** Private flags are in upper word ***/
#define ELF_HIGHLIGHT    0x00010000
#define ELF_SHADOW       0x00020000
#define ELF_COLOUR       0x00040000
#define ELF_FILL         0x00080000
#define ELF_SUNKEN       0x00100000
#define ELF_RAISED       0x00200000
#define ELF_HOLLOW       0X00400000

static ERROR GET_ColourRGB(struct Ellipse *, struct RGB *);
static ERROR GET_Fill(struct Ellipse *, LONG *);
static ERROR GET_Height(struct Ellipse *, struct Variable *);
static ERROR GET_HighlightRGB(struct Ellipse *, struct RGB *);
static ERROR GET_ShadowRGB(struct Ellipse *, struct RGB *);
static ERROR GET_Width(struct Ellipse *, struct Variable *);
static ERROR GET_X(struct Ellipse *, struct Variable *);
static ERROR GET_XOffset(struct Ellipse *, struct Variable *);
static ERROR GET_Y(struct Ellipse *, struct Variable *);
static ERROR GET_YOffset(struct Ellipse *, struct Variable *);

static ERROR SET_Border(struct Ellipse *, STRING);
static ERROR SET_Colour(struct Ellipse *, STRING);
static ERROR SET_Fill(struct Ellipse *, LONG *);
static ERROR SET_Frames(struct Ellipse *, STRING);
static ERROR SET_Highlight(struct Ellipse *, STRING);
static ERROR SET_Hollow(struct Ellipse *, LONG *);
static ERROR SET_Raised(struct Ellipse *, LONG *);
static ERROR SET_Shadow(struct Ellipse *, STRING);
static ERROR SET_Sunken(struct Ellipse *, LONG *);
static ERROR SET_Height(struct Ellipse *, struct Variable *);
static ERROR SET_Width(struct Ellipse *, struct Variable *);
static ERROR SET_X(struct Ellipse *, struct Variable *);
static ERROR SET_XOffset(struct Ellipse *, struct Variable *);
static ERROR SET_Y(struct Ellipse *, struct Variable *);
static ERROR SET_YOffset(struct Ellipse *, struct Variable *);

static struct FieldArray EllipseFields[] = {
   { "Drawable",     0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   { "Frame",        0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   { "Thickness",    0, FDF_LONG|FDF_RW,     0, NULL, NULL },
    /*** Virtual fields ***/
   { "Border",       0, FDF_STRING|FDF_W,    0, NULL,             SET_Border },
   { "Colour",       0, FDF_STRING|FDF_W,    0, NULL,             SET_Colour },
   { "ColourRGB",    0, FDF_POINTER|FDF_R,   0, GET_ColourRGB,    NULL },
   { "Fill",         0, FDF_LONG|FDF_RW,     0, GET_Fill,         SET_Fill },
   { "Frames",       0, FDF_STRING|FDF_W,    0, NULL,             SET_Frames },
   { "Highlight",    0, FDF_STRING|FDF_W,    0, NULL,             SET_Highlight },
   { "HighlightRGB", 0, FDF_POINTER|FDF_R,   0, GET_HighlightRGB, NULL },
   { "Hollow",       0, FDF_LONG|FDF_W,      0, NULL,             SET_Hollow },
   { "Raised",       0, FDF_LONG|FDF_W,      0, NULL,             SET_Raised },
   { "Shadow",       0, FDF_STRING|FDF_W,    0, NULL,             SET_Shadow },
   { "ShadowRGB",    0, FDF_POINTER|FDF_R,   0, GET_ShadowRGB,    NULL },
   { "Sunken",       0, FDF_LONG|FDF_W,      0, NULL,             SET_Sunken },
   /*** Variables ***/
   { "Height",    0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",     0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",    0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",   0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",    0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",   0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

static ERROR ELLIPSE_Free(struct Ellipse *, APTR);
static ERROR ELLIPSE_Init(struct Ellipse *, APTR);
static ERROR ELLIPSE_NewObject(struct Ellipse *, APTR);

static struct ActionArray EllipseActions[] = {
   { AC_Free,          ELLIPSE_Free },
   { AC_Init,          ELLIPSE_Init },
   { AC_NewObject,     ELLIPSE_NewObject },
   { NULL, NULL }
};

static ERROR ReadColour(STRING, LONG *, LONG *, LONG *);

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&EllipseClass, NULL,
      FID_BaseClassID|TLONG, ID_ELLIPSE,
      FID_Version|TFLOAT,    VER_ELLIPSE,
      FID_Name|TSTRING,      "Ellipse",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      EllipseActions,
      FID_Fields|TPTR,       EllipseFields,
      FID_Size|TLONG,        sizeof(struct Ellipse),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (EllipseClass) { Action(AC_Free, EllipseClass, NULL); EllipseClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: DrawEllipse()
*/

static void DrawEllipse(struct Ellipse *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct Variable var;

   /* Determine whether or not we need to draw this ellipse
   ** based on our object's frame settings.
   */

   if (Self->Frame) {
      if (Drawable->Frame != Self->Frame) return;
   }
   else if (Self->Frames) {
      if ((Drawable->Frame > Self->TotalFrames) OR (Self->Frames[Drawable->Frame] < 1)) return;
   }

   /*** Get the object dimensions ***/

   var.Type = FD_DOUBLE;
   GET_X(Self, &var);      Self->Pen->XCoord = var.Double;
   GET_Y(Self, &var);      Self->Pen->YCoord = var.Double;
   GET_Width(Self, &var);  Self->Pen->Width  = var.Double;
   GET_Height(Self, &var); Self->Pen->Height = var.Double;

   Self->Pen->Bitmap = Bitmap;
   Self->Pen->TipSize = Self->Thickness;

   if (Self->Thickness > 0) {
      if (Self->Flags & ELF_HOLLOW) {
         Self->Pen->Red    = Self->ColourRGB.Red;
         Self->Pen->Green  = Self->ColourRGB.Green;
         Self->Pen->Blue   = Self->ColourRGB.Blue;
         Self->Pen->Fill   = TRUE;
         Action(AC_Draw, Self->Pen, NULL);

         Self->Pen->Fill    = FALSE;

         Self->Pen->Red   = Self->Shadow.Red;
         Self->Pen->Green = Self->Shadow.Green;
         Self->Pen->Blue  = Self->Shadow.Blue;
         Action(AC_Draw, Self->Pen, NULL);

         Self->Pen->XCoord--;
         Self->Pen->YCoord--;
         Self->Pen->Red   = Self->Highlight.Red;
         Self->Pen->Green = Self->Highlight.Green;
         Self->Pen->Blue  = Self->Highlight.Blue;
         Action(AC_Draw, Self->Pen, NULL);
      }
      else if (Self->Flags & ELF_RAISED) {

         Self->Pen->Red   = Self->Shadow.Red;
         Self->Pen->Green = Self->Shadow.Green;
         Self->Pen->Blue  = Self->Shadow.Blue;
         Self->Pen->Fill  = FALSE;
         Action(AC_Draw, Self->Pen, NULL);

         Self->Pen->XCoord--;
         Self->Pen->YCoord--;

         Self->Pen->Red   = Self->ColourRGB.Red;
         Self->Pen->Green = Self->ColourRGB.Green;
         Self->Pen->Blue  = Self->ColourRGB.Blue;
         Self->Pen->Fill  = TRUE;
         Action(AC_Draw, Self->Pen, NULL);

         Self->Pen->Red   = Self->Highlight.Red;
         Self->Pen->Green = Self->Highlight.Green;
         Self->Pen->Blue  = Self->Highlight.Blue;
         Self->Pen->Fill  = FALSE;
         Action(AC_Draw, Self->Pen, NULL);
      }
      else if (Self->Flags & ELF_SUNKEN) {
         Self->Pen->Red   = Self->Shadow.Red;
         Self->Pen->Green = Self->Shadow.Green;
         Self->Pen->Blue  = Self->Shadow.Blue;
         Self->Pen->Fill  = FALSE;
         Action(AC_Draw, Self->Pen, NULL);

         Self->Pen->XCoord++;
         Self->Pen->YCoord++;

         Self->Pen->Red   = Self->ColourRGB.Red;
         Self->Pen->Green = Self->ColourRGB.Green;
         Self->Pen->Blue  = Self->ColourRGB.Blue;
         Self->Pen->Fill  = TRUE;
         Action(AC_Draw, Self->Pen, NULL);

         Self->Pen->Red   = Self->Highlight.Red;
         Self->Pen->Green = Self->Highlight.Green;
         Self->Pen->Blue  = Self->Highlight.Blue;
         Self->Pen->Fill  = FALSE;
         Action(AC_Draw, Self->Pen, NULL);
      }
      else {
         Self->Pen->Red    = Self->ColourRGB.Red;
         Self->Pen->Green  = Self->ColourRGB.Green;
         Self->Pen->Blue   = Self->ColourRGB.Blue;
         Self->Pen->Fill   = TRUE;
         Action(AC_Draw, Self->Pen, NULL);

         Self->Pen->Fill    = FALSE;

         /*** Draw the border that needs to surround the ellipse ***/

         Self->Pen->Red   = Self->Highlight.Red;
         Self->Pen->Green = Self->Highlight.Green;
         Self->Pen->Blue  = Self->Highlight.Blue;
         Action(AC_Draw, Self->Pen, NULL);
      }
   }
   else {
      Self->Pen->Red    = Self->ColourRGB.Red;
      Self->Pen->Green  = Self->ColourRGB.Green;
      Self->Pen->Blue   = Self->ColourRGB.Blue;
      Self->Pen->Fill   = TRUE;
      Action(AC_Draw, Self->Pen, NULL);
   }
}

/*****************************************************************************
** Ellipse: Free
*/

static ERROR ELLIPSE_Free(struct Ellipse *Self, APTR Void)
{
   if (Self->DrawableID) {
      OBJECTPTR drawable;
      if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
         ActionTags(MT_SetRenderCallback, drawable, Self, NULL);
         ReleaseObject(drawable);
      }
   }

   if (Self->Pen) {
      Action(AC_Free, Self->Pen, NULL);
      Self->Pen = NULL;
   }

   if (Self->Frames) {
      FreeMemory(Self->Frames);
      Self->Frames = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Ellipse: Init
*/

static ERROR ELLIPSE_Init(struct Ellipse *Self, APTR Void)
{
   OBJECTPTR drawable;

   /*** Find the drawable object that we are associated with ***/

   if (!Self->DrawableID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = OwnerID;
   }

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, drawable, Self, &DrawEllipse);
      ReleaseObject(drawable);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Ellipse: NewObject
*/

static ERROR ELLIPSE_NewObject(struct Ellipse *Self, APTR Void)
{
   if (NewObject(ID_PEN, Self->Head.ObjectFlags|NF_CHILD, &Self->Pen, NULL) IS ERR_Okay) {
      Self->Pen->TipSize  = 1;
      Self->Pen->PenShape = PEN_ELLIPSE;
      Self->Pen->Fill     = FALSE;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_NewObject, ERR_NewObject));
}

/*****************************************************************************
** Function: ReadByte()
** Short:    Reads any double digit hex value and converts it into a number.
*/

static WORD ReadByte(STRING Str)
{
   WORD Value;
   if (!*Str) return(0);
   if ((*Str >= '0') AND (*Str <= '9')) Value = (*Str - '0') * 16;
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value = ((*Str - 'A')+10) * 16;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value = ((*Str - 'a')+10) * 16;
   else Value = 0;
   Str++;
   if ((*Str >= '0') AND (*Str <= '9')) Value += (*Str - '0');
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value += (*Str - 'A') + 10;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value += (*Str - 'a') + 10;
   return(Value);
}


/*****************************************************************************

###FIELD###
Name:   Border
Short:  Set this field to define the colour of the elliptical border.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>To give an ellipse a plain-coloured border, set this field to the
text-based colour value that you want to use (e.g. "255,0,0" for red).
To set the thickness of the border, write a value to the Thickness field.</p>

###SEE ALSO###
Field: Thickness

###END###

*****************************************************************************/

static ERROR SET_Border(struct Ellipse *Self, STRING Value)
{
   SET_Highlight(Self, Value);
   SET_Shadow(Self, Value);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Colour
Short:  Set this field to define the colour of the ellipse.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>The internal colour of an ellipse can be declared by writing to this field.
The colour must be in hexadecimal or separated-decimal format - for example
to create a pure red colour, a setting of "#ff0000" or "255,0,0" would be
valid.</p>

<p>If you create an ellipse without setting the Colour field, the result will
be an unfilled rectangle.  The colour is also intrinsically linked to the Fill
field - if you want to switch the colour off temporarily, you can set the Fill
field to FALSE.</p>

###SEE ALSO###
Fields: ColourRGB, Fill

###END###

*****************************************************************************/

static ERROR SET_Colour(struct Ellipse *Self, STRING Colour)
{
   if (*Colour IS '#') {
      Colour++;
      Self->ColourRGB.Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->ColourRGB.Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->ColourRGB.Blue  = ReadByte(Colour);
   }
   else {
      Self->ColourRGB.Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->ColourRGB.Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->ColourRGB.Blue = StrToInt(Colour);
   }

   Self->Pen->Fill = TRUE;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ColourRGB
Short:  Read this field to get the ellipse colour in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The internal colour of an ellipse can be read in RGB format from this
field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ColourRGB(struct Ellipse *Self, struct RGB *Colour)
{
   Colour->Red   = Self->ColourRGB.Red;
   Colour->Green = Self->ColourRGB.Green;
   Colour->Blue  = Self->ColourRGB.Blue;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  This field can be set to define the drawable area for the ellipse.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>In most cases, when you create an ellipse it should be contained by a
drawable area, such as a Render object.  If you do this correctly then
the ellipse will detect the drawable area and set the Drawable field to the
correct object.  However, there may be occasions where you want to
initialise an ellipse object to a non-graphical container.  If this is the
case then you must set the Drawable field to point to a valid drawable
area, such as a Render object.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Fill
Short:  Set to TRUE to fill the ellipse.
Type:   BOOLEAN
Status: Read/Write

###DESCRIPTION###
<p>If this field is set to TRUE then the ellipse will be filled when it is
drawn.  The colour used to fill the ellipse is determined by the Colour
field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_Fill(struct Ellipse *Self, LONG *Value)
{
   *Value = Self->Pen->Fill;
   return(ERR_Okay);
}

static ERROR SET_Fill(struct Ellipse *Self, LONG *Value)
{
   Self->Pen->Fill = *Value;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Frame
Short:  Limits the drawing of an ellipse graphic to a single frame.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>Setting the Frame field to any value other than zero will force an ellipse
to be drawn only when the drawable's frame matches the specified value.  For
instance, if the drawable container has a frame setting of 2, and the ellipse
has a frame of 1, then no graphics will be drawn to represent the ellipse as
the numbers do not match.</p>

###SEE ALSO###
Field: Frames

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frames
Short:  Determines a sequence of frames that an ellipse may be drawn to.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>For long animation sequences you can set the Frames field to determine
individual frames that an ellipse should be drawn to.  For instance, if a
drawable has been assigned 20 animation frames and you want an ellipse
to appear in frames 5, 6, 7, 8, 17 and 19, you would use this frame
specification:</p>

<pre>
   5-8,17,19
</pre>

<p>Individual frame sets are separated by any non-numeric character such as
a comma, and ranges can be specified through the '-' character.</p>

<p>Setting the Frames field will override any setting that you may have
previously set in the Frame field.</p>

###SEE ALSO###
Field: Frame

###END###

*****************************************************************************/

static ERROR SET_Frames(struct Ellipse *Self, STRING Value)
{
   STRING str;
   LONG frame, end;

   if ((!Value) OR (!(*Value))) return(ERR_Args);

   /*** Calculate the size of the frame array that we are going to need ***/

   Self->TotalFrames = 0;
   str = Value;
   while (*str) {
      if ((*str >= '0') AND (*str <= '9')) {
         frame = StrToInt(str);
         while ((*str >= '0') AND (*str <= '9')) str++;
         if (frame > Self->TotalFrames) Self->TotalFrames = frame;
      }
      else str++;
   }

   Self->TotalFrames++; /* Add an extra frame because we start from #1, not zero */

   /*** Allocate the frame array and record the sequence ***/

   if (AllocMemory((Self->TotalFrames+1) * sizeof(WORD), MEM_DATA, &Self->Frames, NULL) IS ERR_Okay) {
      str = Value;
      while (*str) {
         if ((*str >= '0') AND (*str <= '9')) {
            if ((frame = StrToInt(str)) > 0) {
               while ((*str >= '0') AND (*str <= '9')) str++;
               while ((*str) AND (*str <= 0x20)) str++;
               if (*str IS '-') {
                  while ((*str) AND ((*str < '0') OR (*str > '9'))) str++;
                  if ((end = StrToInt(str)) > 0) {
                     while (frame != end) {
                        Self->Frames[frame] = 1;
                        if (frame < end) frame++;
                        else frame--;
                     }
                     Self->Frames[frame] = 1;
                  }
                  while ((*str >= '0') AND (*str <= '9')) str++;
               }
               else Self->Frames[frame] = 1;
            }
            else while ((*str >= '0') AND (*str <= '9')) str++;
         }
         else str++;
      }

      Self->Frames[Self->TotalFrames] = -1; /* Sequence termination */
   }
   else return(ERR_AllocMemory);

   Self->Frame = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Height
Short:  The height of the ellipse is specified here.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>If the ellipse is to be drawn with a fixed height, set this field to the
desired value.  A percentage may be specified if the FD_PERCENT flag is used
when setting the field.</p>

###SEE ALSO###
Field: Height

###END###

*****************************************************************************/

static ERROR GET_Height(struct Ellipse *Self, struct Variable *Value)
{
   OBJECTPTR Render;
   FLOAT ContainerHeight, YCoord, YOffset;

   if (Value->Type IS FD_DOUBLE) {
      if (Self->Dimensions & DMF_FIXED_HEIGHT) {
         Value->Double = Self->Height;
      }
      else if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->Height * ContainerHeight / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FD_FLOAT, &ContainerHeight);
            GetField(Self, FID_YCoord, FD_FLOAT, &YCoord);
            GetField(Self, FID_YOffset, FD_FLOAT, &YOffset);
            Value->Double = ContainerHeight - YOffset - YCoord;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_FieldNotSet);
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         Value->Double = Self->Height;
      }
      else if (Self->Dimensions & DMF_FIXED_HEIGHT) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->Height / ContainerHeight * 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FD_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_Y) YCoord = Self->YCoord;
            else YCoord = Self->YCoord * 100 / ContainerHeight;
            if (Self->Dimensions & DMF_FIXED_YOFFSET) YOffset = Self->YOffset;
            else YOffset = Self->YOffset * 100 / ContainerHeight;
            Value->Double = (ContainerHeight - YOffset - YCoord) / ContainerHeight * 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_FieldNotSet);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Height(struct Ellipse *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("!SetField:","[Ellipse] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
      Self->Height     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Value->Double <= 0) {
         DPrintF("!SetField:","[Ellipse] A height setting of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
      Self->Height     = Value->Double;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Highlight
Short:  String-based field for setting the highlight colour of a ellipse object.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>If an ellipse object should be highlighted around the edges, set this field
to the desired colour for the highlighting.  The colour specification must be
in the standard hexadecimal or decimal separated format.</p>

<p>Passing a NULL value to this field will turn off highlighting.</p>

###SEE ALSO###
Field: HighlightRGB, Shadow

###END###

*****************************************************************************/

static ERROR SET_Highlight(struct Ellipse *Self, STRING Colour)
{
   if (!Colour) Self->Thickness = 0;
   else {
      ReadColour(Colour, &Self->Highlight.Red, &Self->Highlight.Green, &Self->Highlight.Blue);
      if (!Self->Thickness) Self->Thickness = 1;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   HighlightRGB
Short:  Defines the highlight colour of the ellipse in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The highlight colour of an ellipse can be read in RGB format from this
field.</p>

###SEE ALSO###
Field: Highlight, ShadowRGB, ColourRGB

###END###

*****************************************************************************/

static ERROR GET_HighlightRGB(struct Ellipse *Self, struct RGB *Colour)
{
   Colour->Red   = Self->Highlight.Red;
   Colour->Green = Self->Highlight.Green;
   Colour->Blue  = Self->Highlight.Blue;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Hollow
Short:  If set to TRUE the ellipse will be given a hollow appearance.
Type:   BOOLEAN
Status: Write

###DESCRIPTION###
<p>If you have set the Highlight and Shadow fields of an ellipse object then you
will need to decide whether or not the ellipse should be given a sunken or raised
effect when it is drawn.  Setting the Hollow field to TRUE will give you a
raised appearance and the center of the ellipse will take the form of a
hollow cylinder.</p>

###SEE ALSO###
Field: Sunken, Raised

###END###

*****************************************************************************/

static ERROR SET_Hollow(struct Ellipse *Self, LONG *Value)
{
   if (*Value) {
      if (Self->Flags & ELF_HOLLOW) return(ERR_Okay);
      Self->Flags = (Self->Flags & ~ELF_HOLLOW) | ELF_HOLLOW;
   }
   else Self->Flags &= ~ELF_HOLLOW;
   if (Self->Head.ObjectFlags & NF_INITIALISED) ActionMsg(AC_Draw, Self->DrawableID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Raised
Short:  If set to TRUE the ellipse will appear to be raised into the foreground.
Type:   BOOLEAN
Status: Write

###DESCRIPTION###
<p>If you have set the Highlight and Shadow fields of an ellipse object then you
will need to decide whether or not the ellipse should be given a sunken or raised
effect when it is drawn.  To give it a raised effect you will need to set this
field to TRUE, if not then you should set the Sunken field.</p>

###SEE ALSO###
Field: Sunken

###END###

*****************************************************************************/

static ERROR SET_Raised(struct Ellipse *Self, LONG *Value)
{
   if (*Value) {
      if (Self->Flags & ELF_RAISED) return(ERR_Okay);
      Self->Flags = (Self->Flags & ~ELF_SUNKEN) | ELF_RAISED;
   }
   else Self->Flags &= ~ELF_RAISED;
   if (Self->Head.ObjectFlags & NF_INITIALISED) ActionMsg(AC_Draw, Self->DrawableID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Shadow
Short:  String-based field for setting the shadow colour of an ellipse object.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>If an ellipse object should be shadowed around the edges, set this field
to the desired colour for the shadowing.  The colour specification must be
in the standard hexadecimal or decimal separated format.</p>

<p>Passing a NULL value to this field will turn off shadowing.</p>

###SEE ALSO###
Field: ShadowRGB, Highlight

###END###

*****************************************************************************/

static ERROR SET_Shadow(struct Ellipse *Self, STRING Colour)
{
   if (!Colour) Self->Thickness = 0;
   else {
      ReadColour(Colour, &Self->Shadow.Red, &Self->Shadow.Green, &Self->Shadow.Blue);
      if (!Self->Thickness) Self->Thickness = 1;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ShadowRGB
Short:  Defines the shadow colour of an ellipse in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The shadow colour of an ellipse can be read in RGB format from this field.</p>

###SEE ALSO###
Fields: Shadow, HighlightRGB

###END###

*****************************************************************************/

static ERROR GET_ShadowRGB(struct Ellipse *Self, struct RGB *Value)
{
   Value->Red   = Self->Shadow.Red;
   Value->Green = Self->Shadow.Green;
   Value->Blue  = Self->Shadow.Blue;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Sunken
Short:  Set to TRUE to make an ellipse appear to sink into the background.
Type:   BOOLEAN
Status: Write

###DESCRIPTION###
<p>If you have set the Highlight and Shadow fields of an ellipse object then you
will need to decide whether or not the ellipse should be given a sunken or raised
effect when it is drawn.  To give it a sunken effect you will need to set this
field to TRUE, if not then you should set the Raised field.</p>

###SEE ALSO###
Field: Raised

###END###

*****************************************************************************/

static ERROR SET_Sunken(struct Ellipse *Self, LONG *Value)
{
   if (*Value) {
      if (Self->Flags & ELF_SUNKEN) return(ERR_Okay);
      Self->Flags = (Self->Flags & ~ELF_RAISED) | ELF_SUNKEN;
   }
   else Self->Flags &= ~ELF_SUNKEN;
   if (Self->Head.ObjectFlags & NF_INITIALISED) ActionMsg(AC_Draw, Self->DrawableID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Thickness
Short:  The thickness of the ellipse border is managed by this field.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>The thickness of the ellipse border is managed by this field.  By default
the Thickness value is set to 1, and larger values will inrease the border's
pixel size.  Setting the field to 0 will turn off the border.</p>

###SEE ALSO###
Field: Border

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Width
Short:  The width of the ellipse is specified here.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>If the ellipse is to be drawn with a fixed width, set this field to the
desired value.  A percentage may be specified if the FD_PERCENT flag is used
when setting the field.</p>

###SEE ALSO###
Field: Height

###END###

*****************************************************************************/

static ERROR GET_Width(struct Ellipse *Self, struct Variable *Value)
{
   OBJECTPTR Render;
   FLOAT ContainerWidth, XCoord, XOffset;

   if (Value->Type IS FD_DOUBLE) {
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_WIDTH) {
         Value->Double = Self->Width;
      }
      else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->Width * ContainerWidth / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
            else XCoord = ContainerWidth * Self->XCoord / 100;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) Value->Double = ContainerWidth - XCoord - Self->XOffset;
            else Value->Double = ContainerWidth - XCoord - (ContainerWidth * Self->XOffset / 100);
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_Failed);
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         Value->Double = Self->Width;
      }
      else if (Self->Dimensions & DMF_FIXED_WIDTH) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->Width / ContainerWidth * 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FD_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_X) XCoord = Self->XCoord;
            else XCoord = Self->XCoord * 100 / ContainerWidth;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) XOffset = Self->XOffset;
            else XOffset = Self->XOffset * 100 / ContainerWidth;
            Value->Double = (ContainerWidth - XOffset - XCoord) * 100 / ContainerWidth;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_FieldNotSet);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

static ERROR SET_Width(struct Ellipse *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
      Self->Width      = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
      Self->Width      = Value->Double;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  Defines the horizontal position of the ellipse.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The horizontal position of an ellipse can be fixed to an absolute coordinate
by writing a value to the XCoord field.  If you use the FD_PERCENT flag then
you may write a relative (percentage) based value to the field.  Negative
values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Ellipse *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      OBJECTPTR Render;
      FLOAT ContainerWidth, Width;
      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_X) Value->Double = Self->XCoord;
      else if (Self->Dimensions & DMF_RELATIVE_X) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->XCoord * ContainerWidth / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_WIDTH) Width = Self->Width;
            else Width = ContainerWidth * Self->Width / 100;
            if (Self->Dimensions & DMF_FIXED_XOFFSET) Value->Double = ContainerWidth - Width - Self->XOffset;
            else Value->Double = ContainerWidth - Width - (ContainerWidth * Self->XOffset / 100);
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_Failed);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct Ellipse *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
      Self->XCoord     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
      Self->XCoord     = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  Defines the horizontal offset of the ellipse.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the ellipse will be drawn
from that X coordinate up to the width of the container, minus the value
given in the XOffset.  This means that the width of the Ellipse is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the ellipse will be drawn at an X coordinate calculated from the formula
"XCoord = ContainerWidth - EllipseWidth - XOffset".</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Ellipse *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      OBJECTPTR Render;
      FLOAT ContainerWidth, Width;

      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_XOFFSET) {
         Value->Double = Self->XOffset;
      }
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            Value->Double = Self->XOffset * ContainerWidth / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Width, FT_FLOAT, &ContainerWidth);
            if (Self->Dimensions & DMF_FIXED_WIDTH) Width = Self->Width;
            else Width = ContainerWidth * Self->Width / 100;
            if (Self->Dimensions & DMF_FIXED_X) Value->Double = ContainerWidth - (Self->XCoord + Width);
            else Value->Double = ContainerWidth - ((Self->XCoord * ContainerWidth / 100) + Width);
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_Failed);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Ellipse *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
      Self->XOffset    = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
      Self->XOffset    = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  Defines the vertical position of the ellipse.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The vertical position of an ellipse can be fixed to an absolute coordinate
by writing a value to the YCoord field.  If you use the FD_PERCENT flag then
you may write a relative (percentage) based value to the field.  Negative
values are permitted.</p>

###SEE ALSO###
Field: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct Ellipse *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      OBJECTPTR Render;
      FLOAT ContainerHeight, Height;

      if (Self->Dimensions & DMF_FIXED_Y) {
         Value->Double = Self->YCoord;
      }
      else if (Self->Dimensions & DMF_RELATIVE_Y) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->YCoord * ContainerHeight / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_HEIGHT) Height = Self->Height;
            else Height = ContainerHeight * Self->Height / 100;
            if (Self->Dimensions & DMF_FIXED_YOFFSET) Value->Double = ContainerHeight - Height - Self->YOffset;
            else Value->Double = ContainerHeight - Height - (ContainerHeight * Self->YOffset / 100);
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_Failed);
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {

   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct Ellipse *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
      Self->YCoord     = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
      Self->YCoord     = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  Defines the vertical offset of the ellipse.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the ellipse will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the ellipse is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the ellipse will be drawn at a Y coordinate calculated from the formula
"YCoord = ContainerHeight - EllipseHeight - YOffset".</p>

###SEE ALSO###
Field: YCoord, Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Ellipse *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      OBJECTPTR Render;
      FLOAT ContainerHeight, Height;

      Value->Double = 0;
      if (Self->Dimensions & DMF_FIXED_YOFFSET) {
         Value->Double = Self->YOffset;
      }
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            Value->Double = Self->YOffset * ContainerHeight / 100;
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (AccessObject(Self->DrawableID, 1000, &Render) IS ERR_Okay) {
            GetField(Render, FID_Height, FT_FLOAT, &ContainerHeight);
            if (Self->Dimensions & DMF_FIXED_HEIGHT) Height = Self->Height;
            else Height = ContainerHeight * Self->Height / 100;
            if (Self->Dimensions & DMF_FIXED_Y) Value->Double = ContainerHeight - (Self->YCoord + Height);
            else Value->Double = ContainerHeight - ((Self->YCoord * ContainerHeight / 100) + Height);
            ReleaseObject(Render);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));
      }
      else return(ERR_Failed);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));

   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Ellipse *Self, struct Variable *Value)
{
   if (Value->Type IS FD_DOUBLE) {
      Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
      Self->YOffset    = Value->Double;
   }
   else if (Value->Type IS (FD_DOUBLE|FD_PERCENTAGE)) {
      Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
      Self->YOffset    = Value->Double;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: ReadColour()
*/

static ERROR ReadColour(STRING Colour, LONG *Red, LONG *Green, LONG *Blue)
{
   if (*Colour IS '#') {
      Colour++;
      *Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Blue  = ReadByte(Colour);
   }
   else {
      *Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Blue = StrToInt(Colour);
   }
   return(ERR_Okay);
}
