/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Joyport
Version:   1.0
Status:    Stable
ID:        ID_JOYPORT
Category:  I/O
Date:      May 2001
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 1996-2001.  All rights reserved.
Short:     The Joyport class manages the joyports of the computer.
Keywords:  joystick, joyport, joy, input, joypad, mouse, button

###DESCRIPTION###
<p>The Joyport class provides a common interface that allows you to read data
from joyport hardware.  This class supports both digital and analogue input
from devices that plug into the joyport(s), such as the mouse, joysticks and
joypads.  It manages the data
in a convenient format that is passed through the data channel system.  By
doing so the class also supports software based input in order to support
virtual devices and use of the keyboard to emulate a joystick or mouse for
instance.</p>

<p>New joyport objects can only be created for the purpose of being assigned
to certain hardware or virtual devices.  If you want to create a Joyport object that
represents the default joyport, give it a name of 'SystemJoystick'.  Multiple
joyport objects can be created all sharing the same name, so long as the 
<a href="#Port">Port</a> setting for each object is different.</p>

###END###

*****************************************************************************/

#include <pandora/data/config.h>
#include <pandora/system/all.h>
#include <pandora/input/all.h>

#define JP_MOUSE    0
#define JP_JOYSTICK 1
#define JP_ANALOGUE 2
#define JP_JOYPAD   3
#define JP_KEYBOARD 5

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_JOYPORT, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 1996-2001.  All rights reserved.",
   "May 2001",
   "Joyport"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR JoyClass = 0;

/*****************************************************************************
** Class definition
*/

static struct FieldDef PortTable[] = {
   { "MOUSE1", JPORT_MOUSE1 }, { "MOUSE2", JPORT_MOUSE2 },
   { "MOUSE3", JPORT_MOUSE3 }, { "MOUSE4", JPORT_MOUSE4 },
   { "JOY1",   JPORT_JOY1 },   { "JOY2",   JPORT_JOY2 },
   { "JOY3",   JPORT_JOY3 },   { "JOY4",   JPORT_JOY4 },
   { "JOY5",   JPORT_JOY5 },   { "JOY6",   JPORT_JOY6 },
   { "JOY7",   JPORT_JOY7 },   { "JOY8",   JPORT_JOY8 },
   /*** Synonyms ***/
   { "MOUSE",    JPORT_MOUSE1 },
   { "JOYSTICK", JPORT_JOY1 },
   { "JOYPORT",  JPORT_JOY1 },
   { NULL,     NULL }
};

static struct FieldArray JoyFields[] = {
   { "Port",       0, FDF_LONG|FDF_LOOKUP|FDF_RI, (LONG)&PortTable, NULL, NULL },
   { "XChange",    0, FDF_FLOAT|FDF_R,  0, NULL, NULL },
   { "YChange",    0, FDF_FLOAT|FDF_R,  0, NULL, NULL },
   { "ZChange",    0, FDF_FLOAT|FDF_R,  0, NULL, NULL },
   { "LimitNorth", 0, FDF_FLOAT|FDF_RW, 0, NULL, NULL },
   { "LimitSouth", 0, FDF_FLOAT|FDF_RW, 0, NULL, NULL },
   { "LimitEast",  0, FDF_FLOAT|FDF_RW, 0, NULL, NULL },
   { "LimitWest",  0, FDF_FLOAT|FDF_RW, 0, NULL, NULL },
   END_FIELD
};

static ERROR JOY_DataChannel(struct Joyport *, struct acDataChannel *);
static ERROR JOY_Init(struct Joyport *, APTR);
static ERROR JOY_UserClick(struct Joyport *, struct acUserClick *);
static ERROR JOY_UserClickRelease(struct Joyport *, struct acUserClickRelease *);
static ERROR JOY_UserMovement(struct Joyport *, struct acUserMovement *);

static struct ActionArray JoyActions[] = {
   { AC_DataChannel,      JOY_DataChannel },
   { AC_Init,             JOY_Init },
   { AC_UserClick,        JOY_UserClick },
   { AC_UserClickRelease, JOY_UserClickRelease },
   { AC_UserMovement,     JOY_UserMovement },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&JoyClass, NULL,
      FID_BaseClassID|TLONG, ID_JOYPORT,
      FID_Version|TFLOAT,    VER_JOYPORT,
      FID_Name|TPTR,         "Joyport",
      FID_Category|TLONG,    CCF_IO,
      FID_Actions|TPTR,      JoyActions,
      FID_Fields|TPTR,       JoyFields,
      FID_Size|TLONG,        sizeof(struct Joyport),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (JoyClass) { Action(AC_Free, JoyClass, NULL); JoyClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  DataChannel
Short: This action can be used to send fake joyport input.

###DESCRIPTION###
<p>It is possible to send software generated input to a Joyport object
through this action, as if the user was controlling the device personally.
All data that is sent to a Joyport object through this method will inevitably
be reported to programs that read the Joyport through the DataChannel
service.</p>

<p>Data must be sent using the DATA_JOYPORT type, as described in the Action
List document.</p>

###END###

*****************************************************************************/

static ERROR JOY_DataChannel(struct Joyport *Self, struct acDataChannel *Args)
{
   struct acDataChannel Channel;
   struct ChannelSubscription *list;
   OBJECTID SelfID;
   LONG i;
   MEMORYID ChannelList;

   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((GetField(Self, FID_ChannelList, FT_LONG, &ChannelList) != ERR_Okay) OR (!ChannelList)) return(ERR_Okay);

   if ((Args->DataType IS DATA_JOYPORT) AND (Args->Version IS 1)) {
      /* Send all the data directly to all objects that have
      ** opened data channels with us.
      */

      Channel.ObjectID     = GetUniqueID(Self);
      Channel.DataType     = DATA_JOYPORT;
      Channel.Version      = 1;
      Channel.Buffer       = Args->Buffer;
      Channel.Size         = Args->Size;
      Channel.TotalEntries = Args->TotalEntries;

      SelfID = GetUniqueID(Self);
      if (AccessMemory(ChannelList, MEM_READWRITE, (APTR *)&list) IS ERR_Okay) {
         i = 0;
         while (list[i].SubscriberID) {
            if (ActionMsgPort(AC_DataChannel, list[i].SubscriberID, &Channel, list[i].MessagePortMID, list[i].ClassID) IS ERR_NoMatchingObject) {
               if (UnsubscribeChannel(Self, list[i].SubscriberID) != ERR_Okay) i++;
            }
            else i++;
         }
         ReleaseMemory(list);
      }
   }
   else DPrintF("!DataChannel:","[Joyport] Unrecognised data type %d, version %d.", Args->DataType, Args->Version);

   return(ERR_Okay);
}

/*****************************************************************************
** Joyport: Init
*/

static ERROR JOY_Init(struct Joyport *Self, APTR Void)
{
   if (!Self->Port) Self->Port = JPORT_MOUSE1;

   /* Set the limits.  We also check that the sign of each setting
   ** is correct.
   */

   if (!Self->LimitWest)  Self->LimitWest = -10000.0;
   if (!Self->LimitEast)  Self->LimitEast = -10000.0;
   if (!Self->LimitNorth) Self->LimitNorth = +10000.0;
   if (!Self->LimitSouth) Self->LimitSouth = +10000.0;

   if (Self->LimitWest > 0)  Self->LimitWest = -Self->LimitWest;
   if (Self->LimitEast > 0)  Self->LimitEast = -Self->LimitEast;
   if (Self->LimitNorth < 0) Self->LimitNorth = -Self->LimitNorth;
   if (Self->LimitSouth < 0) Self->LimitSouth = -Self->LimitSouth;

   Self->XChange = 0.0;
   Self->YChange = 0.0;
   Self->ZChange = 0.0;
   return(ERR_Okay);
}

/*****************************************************************************
** Joyport: UserClick
*/

static ERROR JOY_UserClick(struct Joyport *Self, struct acUserClick *Args)
{
   return(ERR_Okay);
}

/*****************************************************************************
** Joyport: UserClickRelease
*/

static ERROR JOY_UserClickRelease(struct Joyport *Self, struct acUserClickRelease *Args)
{
   return(ERR_Okay);
}

/*****************************************************************************
** Joyport: UserMovement
*/

static ERROR JOY_UserMovement(struct Joyport *Self, struct acUserMovement *Args)
{
   struct acDataChannel channel;
   struct dcJoyEntry joy;

   if (!Args) return(ObjectError(Self, ERH_UserMovement, ERR_Args));

   joy.XChange = Args->XChange;
   joy.YChange = Args->YChange;
   joy.ZChange = 0;
   joy.AbsX    = -1;
   joy.AbsY    = -1;
   joy.AbsZ    = -1;
   joy.ButtonPressFlags   = NULL;
   joy.ButtonReleaseFlags = NULL;

   channel.ObjectID     = NULL;
   channel.DataType     = DATA_JOYPORT;
   channel.Version      = 1;
   channel.Buffer       = &joy;
   channel.Size         = sizeof(struct dcJoyEntry);
   channel.TotalEntries = 1;
   Action(AC_DataChannel, Self, &channel);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Names:  LimitEast
Short:  Limits horizontal joyport movement.
Type:   FLOAT
Status: Read/Write

###DESCRIPTION###
<p>The LimitEast field limits the horizontal range for which a joyport device
can send movement values.  If a joyport device sends a value that exceeds a
set limit, the value will be restricted to the maximum value indicated by the
limit setting.  This particular field specifically limits movement to the east
(right) and is paired with the LimitWest field.  Valid values for this field
start from 0 and increase in value to a maximum of 10000.</p>

###SEE ALSO###
Fields: LimitWest, LimitNorth, LimitSouth

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Names:  LimitNorth
Short:  Limits vertical joyport movement.
Type:   FLOAT
Status: Read/Write

###DESCRIPTION###
<p>The LimitNorth field limits the vertical range for which a joyport device
can send movement values.  If a joyport device sends a value that exceeds a
set limit, the value will be restricted to the maximum value indicated by the
limit setting.  This particular field specifically limits movement to the
north (up) and is paired with the LimitSouth field.  Valid values for this
field start from 0 and decrease in value to a maximum of -10000.</p>

###SEE ALSO###
Fields: LimitWest, LimitEast, LimitSouth

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Names:  LimitSouth
Short:  Limits vertical joyport movement.
Type:   FLOAT
Status: Read/Write

###DESCRIPTION###
<p>The LimitSouth field limits the vertical range for which a joyport device
can send movement values.  If a joyport device sends a value that exceeds a
set limit, the value will be restricted to the maximum value indicated by the
limit setting.  This particular field specifically limits movement to the
south (down) and is paired with the LimitNorth field.  Valid values for this
field start from 0 and increase in value to a maximum of 10000.</p>

###SEE ALSO###
Fields: LimitWest, LimitEast, LimitNorth

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Names:  LimitWest
Short:  Limits horizontal joyport movement.
Type:   FLOAT
Status: Read/Write

###DESCRIPTION###
<p>The LimitWest field limits the horizontal range for which a joyport device
can send movement values.  If a joyport device sends a value that exceeds a
set limit, the value will be restricted to the maximum value indicated by the
limit setting.  This particular field specifically limits movement to the west
(left) and is paired with the LimitEast field.  Valid values for this field
start from 0 and decrease in value to a maximum of -10000.</p>

###SEE ALSO###
Fields: LimitEast, LimitNorth, LimitSouth

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:    Port
Short:   The hardware port represented by a Joyport object.
Type:    LONG
Status:  Read/Init

###DESCRIPTION###
<p>This field specifies the port that a Joyport object represents. Generally
this will only be set to 1 (the mouse port) or 2 (the first joystick port).
If set to 3 or higher, the port that is monitored depends on the hardware
configuration.  If there is no hardware associated for a given port, the
Joyport object will be considered to be virtual and will have to receive its
information from an external source.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELDS###
Names:  XChange
Short:  Reflects the most recent change in horizontal movement.
Type:   FLOAT
Status: Read

###DESCRIPTION###
<p>The XChange field reflects the last horizontal value sent from the joyport
device that is represented by the Joyport object.  Please note that it is
considered bad practice to use this field for interpreting joyport information
in real-time, as you will not be able to synchronise your timing with the
field.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELDS###
Names:  YChange
Short:  Reflects the most recent change in vertical movement.
Type:   FLOAT
Status: Read

###DESCRIPTION###
<p>The YChange field reflects the last vertical value sent from the joyport
device that is represented by the Joyport object.  Please note that it is
considered bad practice to use this field for interpreting joyport information
in real-time, as you will not be able to synchronise your timing with the
field.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELDS###
Names:  ZChange
Short:  Reflects the most recent change in inward/outward movement.
Type:   FLOAT
Status: Read/Write

###DESCRIPTION###
<p>The ZChange field reflects the last depth-based (inward/outward) value sent
from the joyport device that is represented by the Joyport object.  This type
of movement often applies to steering wheels that can be pulled out and pushed
in.  Please note that it is considered bad practice to use this field for
interpreting joyport information in real-time, as you will not be able to
synchronise your timing with the field.</p>

###END###

*****************************************************************************/
